import requests
import json
from os.path import isfile
from tqdm import tqdm  # completely optional; we might exclude it


############ Functions used to map IDs from one DB to another
##############################################################
# USAGE:
# generate_mappings(just_pdb_to_uniprot=False) | (dumps to file; checks version and updates automatically; ignores if current)
# PDB_Uniprot("P10970")                        | (checks mappings in the dump files, accepts PDB ID, PDB ID + " " + Chain, or Uniprot ID; no need to specify)


# Help functions
# Download file with replacament if already same filename locally
def download_file(url, output_file, headers=""):
    """Downloads file from the web. Doesn't check for overwrites."""
    with open(output_file, 'w') as new_local_file:
        new_local_file.write(requests.get(url, headers=headers).content.decode('utf-8'))


def json_dump(input_object, output_file):
    """Dumps the input Python object to a .json file using the JSON module. Doesn't check for overwrites."""
    try:
        with open(output_file, 'w') as outf:
            json.dump(input_object, outf)
        return True
    except:
        print("Something went wrong when dumping file")
        return False


def json_load(input_file):
    """Loads a previously dumped Python object from a .json file using the JSON module."""
    try:
        with open(input_file, 'r') as inpf:
            loaded = json.load(inpf)
        return loaded
    except:
        print("Couldn't load.")
        return False


##### PDB <=> Uniprot
# Check if local version of the mapping file is most recent and replace if not
# The returned value says whether the local file is up to date with the remote file
# None means: we don't know because we couldn't fint the online file.
# If no local file we return Fale.
# If you just want to check, best to set no_download to True
def PDB_Uniprot_update_list(local_list_file_path="pdb_chain_uniprot.lst", no_download=False):
    """Two-way mapping between PDB ID and Uniprot IDs.
        1. It checks whether the local version of the mapping file is the most recent and replaces it otherwise.
        2. The returned value (T/F) reports whether the local file is up to date with the remote file.
       If it returns None it means that the online file could not be located. If no local file found, return False.
       Used with the no_download=True option can serve to check version of the file without updating."""
    try:
        # Constant URL for periodically updated PDB<=>Uniprot mappings
        remote_url = "http://ftp.ebi.ac.uk/pub/databases/msd/sifts/text/pdb_chain_uniprot.lst"
        # Load first 100 bytes of the online file to extract its header
        remote_url_headers = {"Range": "bytes=0-100"}
        # Extracting the header (first line)
        remote_file_content = requests.get(remote_url, headers=remote_url_headers).content.decode('utf-8')
        remote_file_header = remote_file_content.split("\n")[0]
    except:
        print("Cannot establish connection. No remote file under this address.")
        return None

    # Check if local file exists
    try:
        # If it does, extract local file's header
        with open(local_list_file_path, 'r') as local:
            local_file_header = local.readline().strip()

        # Compare the local file's header to the remote file's header
        if local_file_header == remote_file_header:
            print("The local mapping file is up to date.")
            return True
        else:
            print("There is a more up to date mapping file available online. Replacing it.")
            if not no_download:
                download_file(remote_url, local_list_file_path)
            return False

    except:
        # No file found. Downloading from remote.
        print("No local file found.")
        if not no_download:
            print("Downloading the remote file")
            download_file(remote_url, local_list_file_path)
        return False


# Parse mapping and put it into a dictionary with "PDBID Chain" key/value pattern
# PDB => Uniprot by default; set false to have it the other way round
def parse_mapping(local_list_file_path="pdb_chain_uniprot.lst", pdb_to_uniprot=True):
    try:
        with open(local_list_file_path, 'r') as raw_mapping:
            # Mapping as a list of lists, each line already divided into columns
            # Already without a header
            raw_mapping = [x.strip().split("\t") for x in raw_mapping.readlines()][1:]
    except:
        print("No local mapping file found. Download it first.")
        return False

    # Create a dictionary with the mapping, depending on the direction
    mapping = dict()
    if pdb_to_uniprot:
        print("Parsing the PDB => Uniprot mapping")
        for line in tqdm(raw_mapping):
            mapping[line[0] + " " + line[1]] = line[2]
    else:
        print("Parsing the Uniprot => PDB mapping")
        for line in tqdm(raw_mapping):
            if line[2] in mapping.keys():
                mapping[line[2]].append(line[0] + " " + line[1])
            else:
                mapping[line[2]] = [line[0] + " " + line[1]]

    return mapping


# Generates mappings if a more recent version can be found and saves them to file
# Returns True if generated anything, returns False if nothing has been generated
def generate_mappings(local_file="pdb_chain_uniprot.lst", just_pdb_to_uniprot=True,
                      pdb_to_uni_mapping_path="pdb_to_uni_mapping.json",
                      uni_to_pdb_mapping_path="uni_to_pdb_mapping.json"):
    """Generates .json dump files containing Python dictionaries with PDB ID=>Uniprot ID and Uniprot ID=>PDB ID mappings."""
    # Check if up to date without downloading first
    if not PDB_Uniprot_update_list(local_list_file_path=local_file, no_download=True):
        # If not, download
        PDB_Uniprot_update_list(local_list_file_path=local_file)

        # Generate new mappings
        json_dump(parse_mapping(pdb_to_uniprot=True), pdb_to_uni_mapping_path)

        if not just_pdb_to_uniprot:
            json_dump(parse_mapping(pdb_to_uniprot=False), uni_to_pdb_mapping_path)

        return True

    # If mapping source file up to date but one of the generated mapping .json files is missing, generate new mapping
    elif not isfile(pdb_to_uni_mapping_path) or not isfile(uni_to_pdb_mapping_path):
        if not isfile(pdb_to_uni_mapping_path):
            json_dump(parse_mapping(pdb_to_uniprot=True), pdb_to_uni_mapping_path)

        if not isfile(uni_to_pdb_mapping_path) and not just_pdb_to_uniprot:
            json_dump(parse_mapping(pdb_to_uniprot=False), uni_to_pdb_mapping_path)

        return True

    else:
        print("Everything is up to date; no need to generate new mapping files.")
        return False


# Look up mapping PDB <=> Uniprot
def PDB_Uniprot(search_key, pdb_to_uniprot_mapping_path="pdb_to_uni_mapping.json",
                uniprot_to_pdb_mapping_path="uni_to_pdb_mapping.json"):
    """Performs lookup in the .json-dumped dictionaries with mappings, both PDB => Uniprot and Uniprot => PDB.
    Search_key can be either a PDBid, a "PDBid Chain" pair or a Uniprot ID; no need to signal it to the function, it recognizes it automatically.
    If search_key == PDBid: return the list of all available mappings
    If search_key == PDBid + " " + Chain: return the only available mapping
    If search_key == UniprotID: return the list of all mappings"""
    if isfile(pdb_to_uniprot_mapping_path):
        try:
            # load PDB => Uniprot mapping
            pdb_to_uni = json_load(pdb_to_uniprot_mapping_path)
        except:
            print("Something went wrong loading the PDB => Uniprot .json mapping file.")
            return False

    if isfile(uniprot_to_pdb_mapping_path):
        try:
            # load Uniprot => PDB mapping
            uni_to_pdb = json_load(uniprot_to_pdb_mapping_path)
        except:
            print("Something went wrong loading the Uniprot => PDB .json mapping file.")

    # If none could be loaded, abort
    if 'pdb_to_uni' not in locals() and 'uni_to_pdb' not in locals():
        print("No mapping could be loaded. Aborting lookup.")
        return False

    # If search_key == PDBid: return the list of all available mappings
    if len(search_key.split()) == 1 and len(search_key) == 4:
        if 'pdb_to_uni' in locals():
            matching_chains = [x for x in pdb_to_uni.keys() if x.startswith(search_key)]
            if len(matching_chains) > 0:
                matches = [tuple([matching_key, pdb_to_uni[matching_key]]) for matching_key in matching_chains]
                return matches
            else:
                print("No mapping found for PDB ID:", search_key)
                return False
        else:
            print("Missing PDB => Uniprot mapping.")
            return False

    # If search_key == PDBid + " " + Chain: return the only available mapping
    elif len(search_key.split()) == 2 and len(search_key.split()[0]) == 4:
        if 'pdb_to_uni' in locals():
            if search_key in pdb_to_uni.keys():
                return pdb_to_uni[search_key]
            else:
                print("No mapping found for:", search_key)
                return False
        else:
            print("Missing PDB => Uniprot mapping.")
            return False

    # If search_key == UniprotID: return the list of all mappings
    elif len(search_key.split()) == 1 and len(search_key) > 4:
        if 'uni_to_pdb' in locals():
            if search_key in uni_to_pdb.keys():
                return uni_to_pdb[search_key]
            else:
                print("No mapping found for Uniprot ID:", search_key)
                return False
        else:
            print("Missing Uniprot => PDB mapping.")
            return False

    # Incorrect search_key
    else:
        print(
            "Incorrect search phrase. Are you sure you're trying to map PDB <=> Uniprot?\nLegal search phrases: 1s72, 1s72 A, P30615")
        return False



from re import compile
import gzip
import bz2


def convertXYZtoPDB(data_xyz, output_gz, start_idx='', stop_idx=''):
    '''Generate a simple PDB file from a XYZ format containing the 3D coordinates of residues
       '''
    converted = ""
    if start_idx != '': start_idx = int(start_idx)
    if stop_idx != '': stop_idx = int(stop_idx)
    if start_idx != '' and stop_idx != '':
        fragment = True
    else:
        fragment = False

    xyz_atm = compile(r"^\s*(?P<resid>[0-9]*)\s+(?P<x>-?\d+\.\d*)\s+(?P<y>-?\d+\.\d*)\s+(?P<z>-?\d+\.\d*).*$")
    for line in data_xyz:
        data = xyz_atm.match(line)
        if data:
            resid = int(data.groups()[0])
            x = float(data.groups()[1])
            y = float(data.groups()[2])
            z = float(data.groups()[3])
            if fragment:
                if resid >= start_idx and resid <= stop_idx:
                    converted += "ATOM%7d  CA  GLY A%4d    %8.3f%8.3f%8.3f  1.00  0.00           C\n" % (
                        resid, resid, x, y, z)
            else:
                converted += "ATOM%7d  CA  GLY A%4d    %8.3f%8.3f%8.3f  1.00  0.00           C\n" % (
                    resid, resid, x, y, z)

    fw = gzip.open(output_gz, 'wb')
    fw.write(converted)
    fw.close()

    return converted


def grepChain(filename, chain, bzip=True):
    atm = compile(r"^ATOM.{9}CA.{6}" + chain + ".*$|^END|^MODEL")
    if bzip:
        f = bz2.BZ2File(filename, "rb")
    else:
        f = open(filename, "r")
    lines = f.readlines()
    f.close()
    out_lines = []

    for line in lines:
        if atm.match(line):
            out_lines.append(line)

    if bzip:
        f = bz2.BZ2File(filename, "wb")
    else:
        f = open(filename, "w")

    f.write("".join(out_lines))
    f.close()


def getSubchain(filename, start_idx, stop_idx, bzip=True, fmt='pdb'):
    if bzip:
        f = bz2.BZ2File(filename, "rb")
        filetype = fmt
    else:
        f = open(filename, "r")
        filetype = fmt

    lines = f.readlines()
    f.close()
    out_lines = []

    if filetype == 'pdb':
        atm = compile(r"^ATOM.{9}CA.*$")
        mdl = compile(r"^END|^MODEL")

        for line in lines:
            if mdl.match(line):
                out_lines.append(line)
            elif atm.match(line):
                i = int(line[22:26])
                if start_idx <= i <= stop_idx:
                    out_lines.append(line)
    elif filetype == 'xyz':
        for line in lines:
            if "t" in line:
                out_lines.append(line)
            else:
                d = line.split()
                i = int(d[0])
                if start_idx <= i <= stop_idx:
                    out_lines.append(line)
    if bzip:
        f = bz2.BZ2File(filename, "wb")
    else:
        f = open(filename, "w")
    f.write("".join(out_lines))
    f.close()



import tarfile
import _pickle
import gzip
import shutil
from os import path, makedirs, listdir

'''Utility methods used by other classes
   '''


def mkDirIfNotExist(dir):
    if not path.exists(dir):
        makedirs(dir)


def mkDirsIfNotExist(dirs):
    for dir in dirs:
        mkDirIfNotExist(dir)


def bz2Files(files, compressed_file, logger=None):
    tar = tarfile.open(compressed_file, "w:bz2")
    for name in files:
        try:
            tar.add(name, path.basename(name))
        except OSError:
            if logger:
                logger.error("Missing file while bz2: " + name)
            else:
                print("Missing file while bz2: " + name)
    tar.close()


def bz2Folder(folder, compressed_file, logger=None):
    mkDirIfNotExist(path.dirname(compressed_file))
    tar = tarfile.open(compressed_file, "w:bz2")
    for name in listdir(folder):
        try:
            tar.add(path.join(folder, name), path.basename(name))
        except OSError:
            if logger:
                logger.error("Missing file while bz2: " + name)
            else:
                print("Missing file while bz2: " + name)
    tar.close()


def arraytostring(ar):
    i = []
    for e in ar:
        i.append("%d-%d" % (e))
    return ", ".join(i)


# Gzip - necessary for JSMOL viewer
def gzipPDB(code, chain, path, suffix=''):
    return _gzip(path + '/' + code + '_' + chain + suffix + '.pdb')


# Gzip
def _gzip(file):
    with open(file, 'rb') as f_in, gzip.open(file + '.gz', 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)


# GUnzip
def _gunzip(file):
    with gzip.open(file, 'rb') as f_in, open(file.replace('.gz', ''), 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)


def read_from_bin_file(filename):
    # Convert digital data to binary format
    with open(filename, 'rb') as file:
        return _pickle.load(file)


def write_file(data, filename):
    # Convert binary data to proper format and write it on Hard Disk
    with open(filename, 'wb') as file:
        file.write(data)



#!/usr/bin/env python3

#                                                                          |--------------|
#    .d8888b.                       8888888b.                  888         | v. alpha 0.3 |
#   d88P  Y88b                      888   Y88b                 888         |  19 Dec 2018 |
#   Y88b.                           888    888                 888         |--------------|
#    "Y888b.   88888b.  888  888    888   d88P 888d888 .d88b.  888888 
#      "Y88b. 888 "88b 888  888     8888888P"  888P"  d88""88b 888    
#         "888 888  888 888  888    888        888    888  888 888    
#   Y88b  d88P 888 d88P Y88b 888    888        888    Y88..88P Y88b.  
#    "Y8888P"  88888P"   "Y88888    888        888     "Y88P"   "Y888 
#              888           888                                   
#              888      Y8b d88P  |------------------------------------------                              
#              888       "Y88P"   | Rapid protein overview from multiple DBs |                              
#   ------------------------------------------------------------------------ |
#   |                            developed                                   |
#   |                               by                                       |
#   |                        Borys Jastrzębski                               |
#   ------------------------------------------------------------------------ |

import wget
from sys import argv
from datetime import datetime
import os
import re


# from prettytable import PrettyTable

###########################################################################################
#                                                                                         #
# To Think/Do: timestamps on result files?                                                #
#              maybe a graphical interface?                                               #
#              function for cleaning up old fastas?                                       #
#              option not to timestamp fastas?                                            #
#              reformat into a table                                                      #
#              collect clan info from the PFAM page                                       #
#              Solve dependency problems on Ubuntu (does the depInstall.sh work?)         #
#              Add option to accessPDB to include only the chain A in the fasta file      #
#              Add knot type distinction (more info needed from KnotProt)                 #
#              Check for multiple chains in PFAM and KnotProt (and also PDB if you don't) #
###########################################################################################

# Input to all: PDB ID

def accessPDB(protID, downloadFasta=True, fastaTimestamp=True):
    # Downloading a temp file to read from
    tempStamp = datetime.now().isoformat()
    try:
        wget.download("https://files.rcsb.org/header/" + protID + ".pdb", ".temps/" + protID + "_" + tempStamp + ".txt")
        print("\nPDB | Database succesfully accessed at ID:", protID)
    except:
        print("PDB | Can't access the database at ID:", protID)
        return None

    # Downloading the fasta sequence
    if downloadFasta:
        # Adding timestamp to the output file based on the preference
        if fastaTimestamp:
            fastaStamp = datetime.now().isoformat()
        else:
            fastaStamp = ""

        # Don't forget to add chain recognition and choice
        # Actually both are stored in one file
        # Ask Lina what to do about it
        try:
            wget.download(
                "https://www.rcsb.org/pdb/download/viewFastaFiles.do?structureIdList=" + protID + "&compressionType=uncompressed",
                "results/fasta/" + protID + "_" + fastaStamp + ".fasta")
            print("\nPDB | The fasta file succesfully downloaded for ID:", protID)
        except:
            print("PDB | Couldn't download the fasta file for ID:", protID)

    # Processing header to extract the remaining data
    with open(".temps/" + protID + "_" + tempStamp + ".txt") as temp:
        # All data to a list of lines
        data = temp.readlines()

    srcOrganism = list(
        (line.split("ORGANISM_SCIENTIFIC:")[1].strip()[:-1] for line in data if "ORGANISM_SCIENTIFIC:" in line))
    protFunction = list((line.split("COMPND")[1].strip() for line in data if "COMPND" in line))

    # Clean up the temporary file
    os.remove(".temps/" + protID + "_" + tempStamp + ".txt")

    return (protID, srcOrganism, protFunction, fastaStamp)


# At the moment it returns a boolean: knotted (in the db) or unknotted (not present in the db)
# WARNING: At the moment it checks only for chain A of the protein.
def accessKnotProt(protID):
    tempStamp = datetime.now().isoformat()
    try:
        # Temporary file in the making
        wget.download("http://knotprot.cent.uw.edu.pl/browse/?pdbid=" + protID,
                      ".temps/" + protID + "_" + tempStamp + ".txt")
        print("\nKnotProt | Database succesfully accessed at ID:", protID)
    except:
        print("KnotProt | Can't access the database at ID:", protID)
        return None

    # Processing the file
    with open(".temps/" + protID + "_" + tempStamp + ".txt") as temp:
        # All data to a list of lines
        data = temp.read()

    # Check the topmost box to see whether knotted or unknotted
    resultBox = data.split('<div class="alert alert-info alert-dismissable">')[1].split("</div>")[0].replace("/",
                                                                                                             "").split(
        "<strong>")

    # Clean up the temporary file
    os.remove(".temps/" + protID + "_" + tempStamp + ".txt")

    if "Knotted" in resultBox:

        tempStamp = datetime.now().isoformat()
        try:
            # Temporary file in the making
            wget.download("http://knotprot.cent.uw.edu.pl/view/" + protID + "/A/",
                          ".temps/" + protID + "_" + tempStamp + ".txt")
            print("\nKnotProt | View succesfully accessed at ID:", protID)
        except:
            print("KnotProt | Can't access the view at ID:", protID)
            return None

        # Processing the file
        with open(".temps/" + protID + "_" + tempStamp + ".txt") as temp:
            # All data to a list of lines
            typeData = temp.read()

        # Check the tag span to check if it's slipknot or knot
        typeData = typeData.split('<span class="label label-primary">')[1].split("</span>")[0]
        # If knotType == True, it's a Slipknot. Else: it's a Knot.
        knotType = ["Knot", "Slipknot"]["Slipknot" in typeData]

        # Clean up the temporary file
        os.remove(".temps/" + protID + "_" + tempStamp + ".txt")

        print("The protein at PDB ID:", protID, "is knotted. It contains a", knotType)
        return knotType

    elif "Unknotted" in resultBox:
        print("The protein at PDB ID:", protID, "is unknotted.")
        return 0
    else:
        print("KnotProt | Error: Can't process the result box for ID:", protID)
        return None


# Takes info for chain A only, and from the PDB part of the PFAM table
# Rewrite to include other chains and then return them to the accessKnotProt function
# To get info on the knots for each chain
def accessPFAM(protID):
    tempStamp = datetime.now().isoformat()
    try:
        # Temporary file in the making
        wget.download("https://pfam.xfam.org/structure/" + protID + "#tabview=tab3",
                      ".temps/" + protID + "_" + tempStamp + ".txt")
        print("\nPFAM | Database succesfully accessed at ID:", protID)
    except:
        print("PFAM | Can't access the database at ID:", protID)
        return None

    with open(".temps/" + protID + "_" + tempStamp + ".txt", 'r') as temp:
        data = temp.read()

    # Clean up the temporary file
    os.remove(".temps/" + protID + "_" + tempStamp + ".txt")

    # Now let's do a simple string split to get the first <tr> of the tbody of the table out of the code.
    # (Only the first row because we're interested in chain A only)
    trOne = "".join(
        data.split('<table id="structuresTable"')[1].split("</table>")[0].split("<tbody>")[1].split("</tr>")[0].split()[
        2:])

    # Then a regex pattern for splitting to get the individual rows of the table
    rows = re.compile(r'<\/?td*>')
    ## Again, the first element will be the name of the chain so we can skip it as it is always A
    table = [x for x in rows.split(trOne) if x != ""][1:]

    # All the relevant information extracted and final values assigned to variables.
    ## The PDB prefix is here in case of potential extension to include seqs starts and ends from UniProt.
    PDBseqStart, PDBseqEnd, famID, famName = table[0], table[1], table[-1].split("family/")[1].split('"')[0], \
                                             table[-1].split(">")[1].split("<")[0]

    # Now we can get the clan info for the family, if it exists
    tempStamp = datetime.now().isoformat()
    try:
        # Temporary file in the making
        wget.download("https://pfam.xfam.org/family/" + famID + "#tabview=tab2",
                      ".temps/" + famID + "_" + tempStamp + ".txt")
        print("\nPFAM | Database succesfully accessed at Family ID:", famID)
    except:
        print("PFAM | Can't access the clan database at Family ID:", famID)
        return None

    with open(".temps/" + famID + "_" + tempStamp + ".txt", 'r') as temp:
        clanData = temp.read()

    # Clean up the temporary file
    os.remove(".temps/" + famID + "_" + tempStamp + ".txt")

    try:
        # Take first <p> of the blockContent div of the clanBlock div
        clanData = \
            clanData.split('<div class="block" id="clanBlock">')[1].split('<div class="blockContent">')[1].split("<p>")[
                1].split("</p>")[0]

        # Simple split to get the name and CL ID from between the <a> handles
        clanID = "CL" + clanData.split("/clan/CL")[1].split('"')[0]
    except:
        print("PFAM | No clan information found for protein family", famID)
        clanID = None

    return (clanID, famID, famName, PDBseqStart, PDBseqEnd)


#########################################################################################################
# SOLVED – THIS IS AN UNNECESSARY FUNCTION AT THE MOMENT. (after you change the display function)       #
# accessPFAM works properly now
# Could serve as a backup                                                                         #
#########################################################################################################
# The mapping file between PDB ID and PFAM comes from ftp://ftp.ebi.ac.uk/pub/databases/Pfam/mappings/. #
# It is a file from 2015 – a bit old, I know, but it's just a temporary measure for the purpose         #
# of this proof of concept.                                                                             #
#########################################################################################################
def mapToPFAM(protID):
    with open(".resources/pdb_pfam_mapping.txt") as mapping:
        data = mapping.readlines()

    header = data[0]
    familyInfo = list((line for line in data if protID.upper() in line))

    if familyInfo == []:
        familyInfo = "No record"

    return (header, familyInfo)


#########################################################################################################
#########################################################################################################

# Deprecated / will be rewritten from scratch
def combineResults(fromPDB, fromKP, fromPFAM):
    # Write out results to a result file
    if fromPDB is not None:

        # Combine everything
        # Write to a result file
        # fromPDB[0] = protID
        with open("results/result_" + fromPDB[0] + ".txt", 'w') as overallOutput:

            # New line in the result file
            def newLine():
                overallOutput.write("\n")

            overallOutput.write("PDB ID of the protein: " + fromPDB[0])
            overallOutput.write("\n--------------------------------------\n\n")

            overallOutput.write("[PDB DATA]")
            newLine()
            overallOutput.write("The protein comes from: " + ", ".join(fromPDB[1]))
            newLine()
            overallOutput.write("Short overview: " + ", ".join(fromPDB[2]))
            overallOutput.write("\n--------------------------------------\n\n")

            overallOutput.write("[KnotProt DATA]")
            newLine()
            overallOutput.write("The protein is: " + ["UNKNOTTED", "KNOTTED"][fromKP])
            overallOutput.write("\n--------------------------------------\n\n")

            overallOutput.write("[PFAM DATA]")
            newLine()
            overallOutput.write(fromPFAM[0])
            overallOutput.write("\n".join(fromPFAM[1]))
            overallOutput.write("\n--------------------------------------\n\n")

            try:
                with open("results/fasta/" + fromPDB[0] + "_" + fromPDB[3] + ".fasta", 'r') as fl:
                    fasta = fl.read()

                overallOutput.write(
                    "[FASTA Sequence from PDB] (results/fasta/" + fromPDB[0] + "_" + fromPDB[3] + ".fasta)")
                newLine()
                overallOutput.write(fasta)
                overallOutput.write("\n--------------------------------------\n\n")
            except:
                print("Trouble finding the fasta file when printing the results. Perhaps nothing was found.")

        print("Result file generated properly. Check the results folder.")
        return "results/result_" + fromPDB[0] + ".txt"

    else:
        print("No protein ID returned by the accessPDB function.")
        return False


def display(filename):
    # display the results file
    try:
        # For MacOS
        os.system("open " + filename)
    except:
        try:
            # For Windows
            os.system("start " + filename)
        except:
            print("Something went wrong displaying the file.")


# Function to clean up older fasta files for one protein
def cleanFastas():
    pass


# Safety measure
def cleanTemps():
    # Clean up all the remaining temporary files
    for temp in [f for f in os.listdir(".temps")]:
        os.remove(".temps/" + temp)


if __name__ == "__main__":
    # Creating necessary folders
    if not os.path.exists(".temps"):
        os.makedirs(".temps")

    if not os.path.exists("../results"):
        os.makedirs("results")

    if not os.path.exists("../results/fasta"):
        os.makedirs("results/fasta")

    # Main loop
    # Since not all of the main functions are ready, this is left commented out
    #
    # for proteinID in argv[1:]:
    #    results = [accessPDB(proteinID), accessKnotProt(proteinID), accessPFAM(proteinID)]
    #    #display(combineResults(*results))
    #    print(results)

    # Extra clean-up
    cleanTemps()



import urllib

import numpy as np

#######
# Author: Pawel Dabrowski-Tumanski
# p.dabrowski@cent.uw.edu.pl
# Date 02.10.2015; version from 28.01.2016
# Script reads the pdb file and gets the chain information. Each chain forms individual class. In each class the coordinates of CA atoms, missing atoms, residues, SS-bonds and links are stored.
# Script converts pdb file to xyz file (one for each chain), in 5-column format (with residue as the last column), fills the gap with straight line and prints warning if the gap is longer than 6 residues.
# Finally it gives the commands for Wanda's program for finding the crossing of the surfaces.
#######

################################ Chain Class definition and possible amino acids ################################
amino_acids = ['ALA', 'ARG', 'ASN', 'ASP', 'CIT', 'CYS', 'GLU', 'GLN', 'GLY', 'HEY', 'HIS', 'HYP', 'ILE', 'LEU', 'LYS',
               'MET', 'ORN', 'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL', 'SEC', 'PYL', 'ASX', 'GLX', 'XLE', 'XAA',
               'MSE', 'FGL', 'LLP', 'SAC', 'PCA', 'MEN', 'CSB', 'HTR', 'PTR', 'SCE', 'M3L', 'OCS', 'KCX', 'SEB', 'MLY',
               'CSW', 'TPO', 'SEP', 'AYA', 'TRN', 'DAL', 'DAR', 'DSG', 'DAS', 'DCY', 'DGN', 'DGL', 'DHI', 'DIL', 'DLE',
               'DLY', 'MED', 'DPN', 'DPR', 'DSN', 'DTH', 'DTR', 'DTY', 'DVA']


def find_index(number, arr):
    for k in range(len(arr)):
        if (arr[k][0] == number):
            return k


class Chain:
    def __init__(self, name):
        self.name = name
        self.residues = []
        self.coordinates = []
        self.bridges = []
        self.missing = []
        self.gaps = []
        self.helix = []
        self.sheet = []

    def find_length(self):
        self.length = max(len(self.residues), len(self.coordinates))

    def add_residue(self, line):
        line = line[19:70]
        for k in range(len(line.split())):
            if (line.split()[k] in amino_acids):
                self.residues.append(line.split()[k])
        self.find_length()

    def add_bridge(self, bridge):
        self.bridges.append(bridge)

    def add_helix(self, helix):
        self.helix.append(helix)

    def add_sheet(self, sheet):
        self.sheet.append(sheet)

    def add_missing(self, missing, residue):
        self.missing.append([missing, residue])

    def add_coordinate(self, index, coordinate, residue):
        if (len(self.coordinates) == 0):
            for k in range(len(self.missing)):
                if (index > self.missing[k][0]):
                    self.coordinates.append([self.missing[k][0], [], self.missing[k][1]])
            self.coordinates.append([index, coordinate, residue])
        else:
            diff = index - self.coordinates[len(self.coordinates) - 1][0]
            if (diff > 1):
                x = self.coordinates[len(self.coordinates) - 1][1][0]
                y = self.coordinates[len(self.coordinates) - 1][1][1]
                z = self.coordinates[len(self.coordinates) - 1][1][2]
                vec_diff = [(coordinate[0] - x) / diff, (coordinate[1] - y) / diff, (coordinate[2] - z) / diff]
                self.gaps.append([diff, self.coordinates[len(self.coordinates) - 1][0], index])
            for k in range(diff - 1):
                x = round(x + vec_diff[0], 3)
                y = round(y + vec_diff[1], 3)
                z = round(z + vec_diff[2], 3)
                #    res=self.missing[find_index(index-diff+k+1,self.missing)][1]
                #    if (res==None):
                res = "XXX"
                self.coordinates.append([index - diff + k + 1, [x, y, z], res])
            self.coordinates.append([index, coordinate, residue])

    ######### characterize the bond type
    def bond_type(self, res1, atom1, Nend, res2, atom2, Cend):
        if ((atom1[0], atom2[0]) == ("C", "N")):
            if ((((res1, atom1) == ("GLU", "CD")) or ((res1, atom1) == ("ASP", "CG"))) and (
                    (res2, atom2) == ("LYS", "NZ"))):
                return "Amide"
            else:
                # return "Amide-like"
                return "Amide"
        if ((atom1[0], atom2[0]) == ("N", "C")):
            if ((((res1, atom1) == ("LYS", "NZ")) or (Nend and atom1 == "N")) and (
                    ((res2, atom2) == ("GLU", "CD")) or ((res2, atom2) == ("ASP", "CG")) or (Cend and atom2 == "C"))):
                return "Amide"
            else:
                # return "Amide-like"
                return "Amide"
        if ((atom1[0], atom2[0]) == ("C", "O")):
            if ((((res1, atom1) == ("GLU", "CD")) or ((res1, atom1) == ("ASP", "CG"))) and (
                    ((res2, atom2) == ("SER", "OG")) or ((res2, atom2) == ("THR", "OG1")))):
                return "Ester"
            else:
                # return "Ester-like"
                return "Ester"
        if ((atom1[0], atom2[0]) == ("O", "C")):
            if ((((res1, atom1) == ("SER", "OG")) or ((res2, atom2) == ("THR", "OG1"))) and (
                    ((res2, atom2) == ("GLU", "CD")) or ((res2, atom2) == ("ASP", "CG")) or (Cend and atom2 == "C"))):
                return "Ester"
            else:
                # return "Ester-like"
                return "Ester"
        if ((atom1[0], atom2[0]) == ("C", "S")):
            if ((((res1, atom1) == ("GLU", "CD")) or ((res1, atom1) == ("ASP", "CG"))) and (
                    (res2, atom2) == ("CYS", "SG"))):
                return "Thioester"
            else:
                # return "Thioester-like"
                return "Thioester"
        if ((atom1[0], atom2[0]) == ("S", "C")):
            if (((res1, atom1) == ("CYS", "SG")) and (
                    ((res2, atom2) == ("GLU", "CD")) or ((res2, atom2) == ("ASP", "CG")) or (Cend and atom2 == "C"))):
                return "Thioester"
            else:
                # return "Thioester-like"
                return "Thioester"
        else:
            if atom1 == 'SG' and atom2 == 'SG':
                return 'SS'
            else:
                return 'Others'

    ######## check, whether it is N- or C-end
    def N_end(self, index):
        if (index == self.coordinates[0][0]):
            return True
        else:
            return False

    def C_end(self, index):
        if (index == self.coordinates[len(self.coordinates) - 1][0]):
            return True
        else:
            return False

    ######### getting taxonomy
    def get_taxonomy(self, name):
        url = "http://www.uniprot.org/uniprot/" + name
        source = urllib.urlopen(url)
        for line in source.decode('utf-8'):
            if ("/taxonomy/33208\"" in line):
                self.taxonomy = "Animal"
                break
            if ("/taxonomy/33090\"" in line):
                self.taxonomy = "Plant"
                break
            if ("/taxonomy/4751\"" in line):
                self.taxonomy = "Fungus"
                break
            if ("/taxonomy/10239\"" in line):
                self.taxonomy = "Virus"
                break
            if ("/taxonomy/2\"" in line):
                self.taxonomy = "Bacteria"
                break
            if ("/taxonomy/2157\"" in line):
                self.taxonomy = "Archea"
                break
            if ("/taxonomy/33630\"" in line):
                self.taxonomy = "Alveolata"
                break
            if ("/taxonomy/554915\"" in line):
                self.taxonomy = "Amoebozoa"
                break
            if ("/taxonomy/554296\"" in line):
                self.taxonomy = "Apusozoa"
                break
            if ("/taxonomy/1401294\"" in line):
                self.taxonomy = "Breviatea"
                break
            if ("/taxonomy/193537\"" in line):
                self.taxonomy = "Centroheliozoa"
                break
            if ("/taxonomy/3027\"" in line):
                self.taxonomy = "Cryptophyta"
                break
            if ("/taxonomy/33682\"" in line):
                self.taxonomy = "Euglenozoa"
                break
            if ("/taxonomy/207245\"" in line):
                self.taxonomy = "Fornicata"
                break
            if ("/taxonomy/38254\"" in line):
                self.taxonomy = "Glaucocystophyceae"
                break
            if ("/taxonomy/2830\"" in line):
                self.taxonomy = "Haptophyceae"
                break
            if ("/taxonomy/5752\"" in line):
                self.taxonomy = "Heterolobosea"
                break
            if ("/taxonomy/556282\"" in line):
                self.taxonomy = "Jakobida"
                break
            if ("/taxonomy/339960\"" in line):
                self.taxonomy = "Katablepharidophyta"
                break
            if ("/taxonomy/136087\"" in line):
                self.taxonomy = "Malawimonadidae"
                break
            if ("/taxonomy/66288\"" in line):
                self.taxonomy = "Oxymonadida"
                break
            if ("/taxonomy/5719\"" in line):
                self.taxonomy = "Parabasalia"
                break
            if ("/taxonomy/543769\"" in line):
                self.taxonomy = "Rhizaria"
                break
            if ("/taxonomy/2763\"" in line):
                self.taxonomy = "Rhodophyta"
                break
            if ("/taxonomy/33634\"" in line):
                self.taxonomy = "Stramenopiles"
                break
            if ("/taxonomy/33630\"" in line):
                self.taxonomy = "Alveolata"
                break
            else:
                self.taxonomy = "Other"

    ######### cleaning data
    def clean(self):
        ### adding C-end
        if (len(self.coordinates) != 0):
            for k in range(len(self.missing)):
                if (self.coordinates[len(self.coordinates) - 1][0] < self.missing[k]):
                    self.coordinates.append([self.missing[k][0], [], self.missing[k][1]])
        else:
            for k in range(len(self.missing)):
                self.coordinates.append([self.missing[k][0], [], self.missing[k][1]])
                ### clearing double CA atoms
        for k in range(len(self.coordinates) - 1, 0, -1):
            if (self.coordinates[k][0] == self.coordinates[k - 1][0]):
                self.coordinates.pop(k)
        #   if (self.coordinates_residue[k][0]==self.coordinates[k-1][0]):
        #    self.coordinates_residue.pop(k)
        ### clearing non-protein links
        for k in range(len(self.bridges) - 1, -1, -1):
            if ((self.bridges[k][1] not in amino_acids) or (self.bridges[k][4] not in amino_acids)):
                self.bridges.pop(k)
        ### clearing to small loops
        for k in range(len(self.bridges) - 1, -1, -1):
            if (abs(self.bridges[k][3] - self.bridges[k][6]) < 5):
                self.bridges.pop(k)
        ### defining bond type
        for k in range(len(self.bridges)):
            if (self.bridges[k][0] == "LINK"):
                self.bridges[k][0] = self.bond_type(self.bridges[k][1], self.bridges[k][2],
                                                    self.N_end(self.bridges[k][3]), self.bridges[k][4],
                                                    self.bridges[k][5], self.C_end(self.bridges[k][6]))
        ### removing bonds and links which do not exist!
        for k in range(len(self.bridges) - 1, -1, -1):
            if (find_index(self.bridges[k][3], self.coordinates) == None or find_index(self.bridges[k][6],
                                                                                       self.coordinates) == None):
                self.bridges.pop(k)
        ### removing amide bonds used to enlarge the backbone
        for k in range(len(self.bridges) - 1, -1, -1):
            # if ((self.bridges[k][0]=="Amide-like") and (abs(find_index(self.bridges[k][3],self.coordinates)-find_index(self.bridges[k][6],self.coordinates))==1)):
            if ((self.bridges[k][0] == "Amide") and (
                    abs(find_index(self.bridges[k][3], self.coordinates) - find_index(self.bridges[k][6],
                                                                                      self.coordinates)) == 1)):
                self.bridges.pop(k)
        ### dealing with different number of residues in different parts of PDB file
        k = len(self.residues) - len(self.coordinates)
        if (k > 0):
            check = 1  # check=1, if residues in N-terminus are same
            for i in range(min(k, len(self.residues), len(self.coordinates))):
                if (self.residues[i] != self.coordinates[i][2]):
                    check = 0
            if (check == 0):
                for i in range(k):
                    self.coordinates.insert(0, [self.coordinates[0][0] - 1, [], ""])
            else:
                if (len(self.residues) != 0 and len(self.coordinates) != 0):
                    for i in range(len(self.residues) - k, len(self.residues), 1):
                        self.coordinates.append([self.coordinates[len(self.coordinates) - 1][0] + 1, [], ""])
        if (k < 0):
            check = 1  # check=1, if residues in N-terminus are same
            for i in range(k):
                if (self.residues[i] != self.coordinates[i][2]):
                    check = 0
            if (check == 0):
                for i in range(k):
                    self.residues.insert(0, "UNK")
            else:
                for i in range(len(self.residues), len(self.residues) + k, 1):
                    self.residues.append("UNK")
                    ### length check, just for sure
        self.find_length()
        ######### checking gaps

    def check_gaps(self):
        communicate = "\n"
        for k in range(len(self.gaps)):
            if (self.gaps[k][0] > 7):
                communicate += "WARNING!!! In chain " + self.name + " there is a gap of length " + str(
                    self.gaps[k][0] - 1) + " between residues " + str(self.gaps[k][1]) + " and " + str(
                    self.gaps[k][2]) + "\n"
        communicate = communicate[:-1]
        return communicate

    ######### printing data
    def chain_print(self, PDB):
        output_file = open(PDB + "_" + self.name + ".xyz", 'w')
        print(self.check_gaps())
        if (len(self.residues) == self.length):
            for k in range(min(self.length, len(self.coordinates))):
                if (self.coordinates[k][1] != []):
                    output_file.write(str(self.coordinates[k][0]) + " " + str(self.coordinates[k][1][0]) + " " + str(
                        self.coordinates[k][1][1]) + " " + str(self.coordinates[k][1][2]) + " " + str(
                        self.residues[k]) + "\n")
        else:
            for k in range(self.length):
                if (self.coordinates[k][1] != []):
                    output_file.write(str(self.coordinates[k][0]) + " " + str(self.coordinates[k][1][0]) + " " + str(
                        self.coordinates[k][1][1]) + " " + str(self.coordinates[k][1][2]) + " " + str(
                        self.coordinates[k][2]) + "\n")
        output_file.close()
        output_file = open(PDB + "_" + self.name + ".pdb", 'w')
        for k in range(len(self.helix)):
            output_file.write(self.helix[k])
        for k in range(len(self.sheet)):
            output_file.write(self.sheet[k])
        if (len(self.residues) == self.length):
            for k in range(min(self.length, len(self.coordinates))):
                if (self.coordinates[k][1] != []):
                    output_file.write(
                        "ATOM  %(atom)5s  CA  %(resname)3s A%(res)4s    %(x)8s%(y)8s%(z)8s  1.00  1.00           C\n" % {
                            "atom": self.coordinates[k][0], "res": self.coordinates[k][0], "resname": self.residues[k],
                            "x": self.coordinates[k][1][0], "y": self.coordinates[k][1][1],
                            "z": self.coordinates[k][1][2]})
        else:
            for k in range(self.length):
                if (self.coordinates[k][1] != []):
                    output_file.write(
                        "ATOM  %(atom)5s  CA  %(resname)3s A%(res)4s    %(x)8s%(y)8s%(z)8s  1.00  1.00           C\n" % {
                            "atom": self.coordinates[k][0], "res": self.coordinates[k][0],
                            "resname": self.coordinates[k][2], "x": self.coordinates[k][1][0],
                            "y": self.coordinates[k][1][1], "z": self.coordinates[k][1][2]})
        output_file.write("END\n")
        output_file.close()

    ######### printing commands to Wanda's program
    def commands_print(self, PDB, flag=0):
        for k in range(len(self.bridges)):
            if (flag == 1):
                print(self.bridges[k][0] + " ./surfacesMyOrient " + PDB + "_" + self.name + ".xyz " + str(
                    self.bridges[k][3]) + " " + str(self.bridges[k][6]) + " 0 0")
            if (flag == 2):
                print(self.bridges[k][0] + " " + PDB + "_" + self.name + " " + str(self.bridges[k][3]) + " " + str(
                    self.bridges[k][6]))
            else:
                print("./surfacesMyOrient " + PDB + "_" + self.name + ".xyz " + str(self.bridges[k][3]) + " " + str(
                    self.bridges[k][6]) + " 0 0")

    def getBridges(self, filepath):
        result = []
        for k in range(len(self.bridges)):
            result.append((
                self.bridges[k][0],
                self.name.upper(),
                self.bridges[k][3],
                self.bridges[k][6],
                # "./surfacesMyOrient "+filepath+"_"+self.name+".xyz "+str(self.bridges[k][3])+" "+str(self.bridges[k][6])+" 0 0",
            ))
        return result

    def getGaps(self):
        first = True
        result = ''
        for g in self.gaps:
            if not first:
                result += ', '
            first = False
            result += str(g[1]) + '-' + str(g[2])
        return result

    ######### finding distance between residues
    def find_distance(self, res1, res2):
        for k in range(len(self.coordinates)):
            if (self.coordinates[k][0] == res1):
                vec1 = self.coordinates[k][1]
            if (self.coordinates[k][0] == res2):
                vec2 = self.coordinates[k][1]
        print(np.linalg.norm(np.asarray(vec1) - np.asarray(vec2)))

    ######### finding index of first residue with coordinates
    def find_first(self):
        for k in range(len(self.coordinates)):
            if (self.coordinates[k][1] != []):
                return self.coordinates[k][0]


################################ Main part ################################
### search for chains and build chain classes
def run_pdb2Wanda(infile, work_dir, filename):
    with open(infile, 'r') as f:
        lines = []
        for line in f.readlines():
            lines.append(line.strip('\n'))

    chains = []
    for line in lines:
        if (line[0:6] == "SEQRES"):
            if (line[11] not in chains):
                chains.append(line[11])
        if (line[0:4] == "ATOM"):
            if (line[21] not in chains):
                chains.append(line[21])

    ### fill the chain class information
    cross_chain_bridges = []
    result = []
    result_gaps = {}
    result_str_begin = {}
    chain_data = {}
    for k in range(len(chains)):
        chain_data[chains[k]] = Chain(chains[k])
        for line in lines:
            if ((line[0:10] == "REMARK 465") and (line[15:18] in amino_acids) and (
                    line[19] == chains[k]) and isinstance(line[21:26], int)):
                chain_data[chains[k]].add_missing(int(line[21:26]), line[15:18])
            #  if ((line[0:5]=="DBREF") and (line[26:32].strip()=="UNP") and (line[12]==chains[k])):	# gets the taxonomy, but slows down the program
            #   chain_data[chains[k]].get_taxonomy(line[33:41].strip())
            if ((line[0:6] == "SEQRES") and (line[11] == chains[k])):
                chain_data[chains[k]].add_residue(line)
            if ((line[0:5] == "HELIX") and (line[19] == chains[k]) and (line[31] == chains[k])):
                chain_data[chains[k]].add_helix(line)
            if ((line[0:5] == "SHEET") and (line[21] == chains[k]) and (line[32] == chains[k])):
                chain_data[chains[k]].add_sheet(line)
            if ((line[0:6] == "SSBOND") and (line[15] == chains[k]) and (line[29] == chains[k])):
                chain_data[chains[k]].add_bridge(
                    ["SS", line[11:14], "S", int(line[17:21]), line[25:28], "S", int(line[31:35])])
            if ((line[0:6] == "SSBOND") and (line[15] != line[29])):
                cross_chain_bridges.append(
                    ["SS", line[15], line[11:14], "S", int(line[17:21]), line[29], line[25:28], "S", int(line[31:35])])
            if ((line[0:4] == "LINK") and (line[21] == chains[k]) and (line[51] == chains[k])):
                chain_data[chains[k]].add_bridge(
                    ["LINK", line[17:20], line[12:16].strip(), int(line[22:26]), line[47:50], line[42:46].strip(),
                     int(line[52:56])])
            if ((line[0:4] == "LINK") and (line[21] != line[51])):
                cross_chain_bridges.append(
                    ["LINK", line[21], line[17:20], line[12:16].strip(), int(line[22:26]), line[51], line[47:50],
                     line[42:46].strip(), int(line[52:56])])
            if ((line[0:6] == "HETATM") and (line[13:15] == "CA") and (line[21] == chains[k])):
                chain_data[chains[k]].add_coordinate(int(line[22:26]),
                                                     [float(line[30:38]), float(line[38:46]), float(line[46:54])],
                                                     line[17:20])
            if ((line[0:4] == "ATOM") and (line[13:15] == "CA") and (line[21] == chains[k])):
                chain_data[chains[k]].add_coordinate(int(line[22:26]),
                                                     [float(line[30:38]), float(line[38:46]), float(line[46:54])],
                                                     line[17:20])
            if ((line[0:3] == "TER") and (len(chain_data[chains[k]].coordinates) != 0)):
                break
        chain_data[chains[k]].clean()  # clean bridges data - do not comment
        chain_data[chains[k]].chain_print(work_dir + filename)  # save coordinates to .xyz and .pdb file
        # chain_data[chains[k]].commands_print(sys.argv[1])	# print input commands to Wanda's program consistent with files saved line above. Addition of second argument=1 prints as first field the bond type
        result.extend(chain_data[chains[k]].getBridges(work_dir + filename))
        result_gaps[chain_data[chains[k]].name] = chain_data[chains[k]].getGaps()
        result_str_begin[chain_data[chains[k]].name] = chain_data[chains[k]].find_first()
    return (result, result_gaps, result_str_begin)



from .Command import *
from .cddf_parser import *
from .common import *
from .EditPdbXyz import *
from .fetchPDBinfo import *
from .fetchChainInfo import *
from .fetchAnnotations import *
from .idMapping import *
from .spyProt import *



import numpy as np
import itertools
from Bio.PDB.MMCIF2Dict import MMCIF2Dict

#######
# Modified by Bartosz Gren, b.gren@cent.uw.edu.pl
# Author: Pawel Dabrowski-Tumanski
# p.dabrowski@cent.uw.edu.pl
# Date 02.10.2015; version from 28.01.2016
# Script reads the pdb file and gets the chain information. Each chain forms 
#   individual class. In each class the coordinates of CA atoms, missing atoms,
#   residues, SS-bonds and links are stored.
# Script converts pdb file to xyz file (one for each chain), in 5-column format
#   (with residue as the last column), fills the gap with straight line and 
#   prints warning if the gap is longer than 6 residues.
# Finally it gives the commands for Wanda's program for finding the crossing of
#   the surfaces.
#######

############ Chain Class definition and possible amino acids ##################
AMINOACIDS = ['ALA', 'ARG', 'ASN', 'ASP', 'CIT', 'CYS', 'GLU', 'GLN', 'GLY', 'HEY', 'HIS',
              'HYP', 'ILE', 'LEU', 'LYS', 'MET', 'ORN', 'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL',
              'SEC', 'PYL', 'ASX', 'GLX', 'XLE', 'XAA', 'MSE', 'FGL', 'LLP', 'SAC', 'PCA', 'MEN', 'CSB',
              'HTR', 'PTR', 'SCE', 'M3L', 'OCS', 'KCX', 'SEB', 'MLY', 'CSW', 'TPO', 'SEP', 'AYA', 'TRN',
              'DAL', 'DAR', 'DSG', 'DAS', 'DCY', 'DGN', 'DGL', 'DHI', 'DIL', 'DLE', 'DLY', 'MED', 'DPN',
              'DPR', 'DSN', 'DTH', 'DTR', 'DTY', 'DVA']


def find_index(number, arr):
    for k in range(len(arr)):
        if (arr[k][0] == number):
            return k


class Chain:
    def __init__(self, name):
        self.name = name
        self.residues = []
        self.coordinates = []
        self.bridges = []
        self.missing = []
        self.gaps = []
        self.helix = []
        self.sheet = []
        self.length = None

    def find_length(self):
        self.length = max(len(self.residues), len(self.coordinates))

    def add_residue(self, resid):
        if resid in AMINOACIDS:
            self.residues.append(resid)
        self.find_length()  # relikt z pdb2wanda

    def add_bridge(self, bridge):
        self.bridges.append(bridge)

    def add_helix(self, helix):
        self.helix.append(helix)

    def add_sheet(self, sheet):
        self.sheet.append(sheet)

    def add_missing(self, missing, residue):
        self.missing.append([missing, residue])

    def add_coordinate(self, index, coordinate, residue):
        if (len(self.coordinates) == 0):
            for k in range(len(self.missing)):
                if (index > self.missing[k][0]):
                    self.coordinates.append([self.missing[k][0], [], self.missing[k][1]])
            self.coordinates.append([index, coordinate, residue])
        else:
            diff = index - self.coordinates[len(self.coordinates) - 1][0]
            if (diff > 1):
                x = self.coordinates[len(self.coordinates) - 1][1][0]
                y = self.coordinates[len(self.coordinates) - 1][1][1]
                z = self.coordinates[len(self.coordinates) - 1][1][2]
                vec_diff = [(coordinate[0] - x) / diff, (coordinate[1] - y) / diff, (coordinate[2] - z) / diff]
                self.gaps.append([diff, self.coordinates[len(self.coordinates) - 1][0], index])
            for k in range(diff - 1):
                x = round(x + vec_diff[0], 3)
                y = round(y + vec_diff[1], 3)
                z = round(z + vec_diff[2], 3)
                #                res=self.missing[find_index(index-diff+k+1,self.missing)][1]
                #                if (res==None):
                res = "XXX"
                self.coordinates.append([index - diff + k + 1, [x, y, z], res])
            self.coordinates.append([index, coordinate, residue])

    ######### characterize the bond type 
    def bond_type(self, res1, atom1, Nend, res2, atom2, Cend):
        a1, a2 = sorted((atom1[0], atom2[0]))
        if atom1 == atom2 == 'SG':
            return 'SS'
        elif a1 == "C":
            if a2 == "N":
                return "Amide"
            elif a2 == "O":
                return "Ester"
            elif a2 == "S":
                return "Thioester"
        return 'Others'

    ######## check, whether it is N- or C-end
    def N_end(self, index):
        return index == self.coordinates[0][0]

    def C_end(self, index):
        return index == self.coordinates[len(self.coordinates) - 1][0]

    ######### cleaning data
    def clean(self):
        ### adding C-end
        if len(self.coordinates) != 0:
            for missing in self.missing:
                if (self.coordinates[len(self.coordinates) - 1][0] < missing[0]):
                    self.coordinates.append([missing[0], [], missing[1]])
        else:
            for missing in self.missing:
                self.coordinates.append([missing[0], [], missing[1]])
                ### clearing double CA atoms
        for k in range(len(self.coordinates) - 1, 0, -1):
            if (self.coordinates[k][0] == self.coordinates[k - 1][0]):
                self.coordinates.pop(k)
        ### clearing non-protein links
        for k in range(len(self.bridges) - 1, -1, -1):
            if ((self.bridges[k][1] not in AMINOACIDS) or (self.bridges[k][4] not in AMINOACIDS)):
                self.bridges.pop(k)
        ### clearing to small loops
        for k in range(len(self.bridges) - 1, -1, -1):
            if (abs(self.bridges[k][3] - self.bridges[k][6]) < 5):
                self.bridges.pop(k)
        ### defining bond type -- now in main
        #for bridge in self.bridges:
        #    if (bridge[0] == "LINK"):
        #        bridge[0] = self.bond_type(bridge[1], bridge[2], self.N_end(bridge[3]),
        #                                   bridge[4], bridge[5], self.C_end(bridge[6]))
        ### removing bonds and links which do not exist!
        for k in range(len(self.bridges) - 1, -1, -1):
            if (find_index(self.bridges[k][3], self.coordinates) == None
                    or find_index(self.bridges[k][6], self.coordinates) == None):
                self.bridges.pop(k)
        ### removing amide bonds used to enlarge the backbone
        for k in range(len(self.bridges) - 1, -1, -1):
            index_diff = find_index(self.bridges[k][3], self.coordinates) \
                         - find_index(self.bridges[k][6], self.coordinates)
            if self.bridges[k][0] == "Amide" and abs(index_diff) == 1:
                self.bridges.pop(k)
        ### dealing with different number of residues in different parts of PDB file
        k = len(self.residues) - len(self.coordinates)
        if k > 0:
            check = 1  # check=1, if residues in N-terminus are same
            for i in range(min(k, len(self.residues), len(self.coordinates))):
                if (self.residues[i] != self.coordinates[i][2]):
                    check = 0
            if check == 0:
                for i in range(k):
                    self.coordinates.insert(0, [self.coordinates[0][0] - 1, [], ""])
            else:
                if len(self.residues) != 0 and len(self.coordinates) != 0:
                    for i in range(len(self.residues) - k, len(self.residues), 1):
                        self.coordinates.append([self.coordinates[len(self.coordinates) - 1][0] + 1, [], ""])
        if k < 0:
            check = 1  # check=1, if residues in N-terminus are same
            for i in range(k):
                if (self.residues[i] != self.coordinates[i][2]):
                    check = 0
            if check == 0:
                for i in range(k):
                    self.residues.insert(0, "UNK")
            else:
                for i in range(len(self.residues), len(self.residues) + k, 1):
                    self.residues.append("UNK")
                    ### length check, just for sure
        self.find_length()

        ######### checking gaps

    def check_gaps(self):
        communicate = ""
        for gap in self.gaps:
            if gap[0] > 7:
                communicate += "\nWARNING!!! In chain {} there is a gap of " \
                               "length {} between residues {} and {}\n".format( \
                    self.name, str(gap[0] - 1), str(gap[1]), gap[2])
        return communicate

    ######### printing data
    def chain_print(self, PDB):
        output_file = open(PDB + "_" + self.name + ".xyz", 'w')
        print(self.check_gaps())
        if (len(self.residues) == self.length):
            for k in range(min(self.length, len(self.coordinates))):
                if (self.coordinates[k][1] != []):
                    output_file.write(str(self.coordinates[k][0]) + " " + str(self.coordinates[k][1][0]) + " " + str(
                        self.coordinates[k][1][1]) + " " + str(self.coordinates[k][1][2]) + " " + str(
                        self.residues[k]) + "\n")
        else:
            for k in range(self.length):
                if (self.coordinates[k][1] != []):
                    output_file.write(str(self.coordinates[k][0]) + " " + str(self.coordinates[k][1][0]) + " " + str(
                        self.coordinates[k][1][1]) + " " + str(self.coordinates[k][1][2]) + " " + str(
                        self.coordinates[k][2]) + "\n")
        output_file.close()
        output_file = open(PDB + "_" + self.name + ".pdb", 'w')
        for k in range(len(self.helix)):
            output_file.write(self.helix[k])
        for k in range(len(self.sheet)):
            output_file.write(self.sheet[k])
        if (len(self.residues) == self.length):
            for k in range(min(self.length, len(self.coordinates))):
                if (self.coordinates[k][1] != []):
                    output_file.write(
                        "ATOM  %(atom)5s  CA  %(resname)3s A%(res)4s    %(x)8s%(y)8s%(z)8s  1.00  1.00           C\n" % {
                            "atom": self.coordinates[k][0], "res": self.coordinates[k][0], "resname": self.residues[k],
                            "x": self.coordinates[k][1][0], "y": self.coordinates[k][1][1],
                            "z": self.coordinates[k][1][2]})
        else:
            for k in range(self.length):
                if (self.coordinates[k][1] != []):
                    output_file.write(
                        "ATOM  %(atom)5s  CA  %(resname)3s A%(res)4s    %(x)8s%(y)8s%(z)8s  1.00  1.00           C\n" % {
                            "atom": self.coordinates[k][0], "res": self.coordinates[k][0],
                            "resname": self.coordinates[k][2], "x": self.coordinates[k][1][0],
                            "y": self.coordinates[k][1][1], "z": self.coordinates[k][1][2]})
        output_file.write("END\n")
        output_file.close()

    ######### printing commands to Wanda's program
    def commands_print(self, PDB, flag=0):
        for k in range(len(self.bridges)):
            if (flag == 1):
                print(self.bridges[k][0] + " ./surfacesMyOrient " + PDB + "_" + self.name + ".xyz " + str(
                    self.bridges[k][3]) + " " + str(self.bridges[k][6]) + " 0 0")
            if (flag == 2):
                print(self.bridges[k][0] + " " + PDB + "_" + self.name + " " + str(self.bridges[k][3]) + " " + str(
                    self.bridges[k][6]))
            else:
                print("./surfacesMyOrient " + PDB + "_" + self.name + ".xyz " + str(self.bridges[k][3]) + " " + str(
                    self.bridges[k][6]) + " 0 0")

    def getBridges(self):
        result = []
        for bridge in self.bridges:
            result.append((bridge[0], self.name.upper(), bridge[3], bridge[6]))
        return result

    def getGaps(self):
        first = True
        result = ''
        for g in self.gaps:
            if not first:
                result += ', '
            first = False
            result += str(g[1]) + '-' + str(g[2])
        return result

    ######### finding distance between residues
    def find_distance(self, res1, res2):
        for k in range(len(self.coordinates)):
            if (self.coordinates[k][0] == res1):
                vec1 = self.coordinates[k][1]
            if (self.coordinates[k][0] == res2):
                vec2 = self.coordinates[k][1]
        print(np.linalg.norm(np.asarray(vec1) - np.asarray(vec2)))

    ######### finding index of first residue with coordinates
    def find_first(self):
        for k in range(len(self.coordinates)):
            if (self.coordinates[k][1] != []):
                return self.coordinates[k][0]


##########################################################################
### functions loading cif data item columns, please use them instead of "cifdict"

# cifdict returns list of strings or a string, this function
# unites this behaviour and always returns a list.
def get_feature(feature, cifdict):
    data = cifdict[feature]
    if type(data) == str:
        return [data]
    return data

# returns generator which generates lines over a zipped lists of features
def iterate_with_features(features, cifdict):
    if type(features) == str:
        features = [features]
    data = [get_feature(feature, cifdict) for feature in features]
    for line in zip(data):
        yield line

################################ Main part ################################
### search for chains and build chain classes
def run_cif2Wanda(infile, work_dir, outfile):
    # ignored search for secondary structures
    # infile -- file name or handle
    cifdict = MMCIF2Dict(infile) # in the rest of code please do not use "cifdict"
                                 # use "get_features" or "iterate_with_features"
    # features, all names can be checked here https://mmcif.wwpdb.org/pdbx-mmcif-home-page.html
    f_chains = '_struct_asym.id' # chain
    f_entities = '_entity.id'    # entity (unique chain)

    # features of sequence
    f_seq_chain = '_pdbx_poly_seq_scheme.asym_id'       # asymmetric unit
    f_seq_resid = '_pdbx_poly_seq_scheme.mon_id'        # residue
    f_seq_resid_pdb ='_pdbx_poly_seq_scheme.pdb_mon_id' # residue according to pdb
    f_seq_seqid = '_pdbx_poly_seq_scheme.seq_id'        # sequence id
    f_seq_list = [f_seq_chain, f_seq_resid, f_seq_resid_pdb, f_seq_seqid]

    # features of links
    f_link_type = '_struct_conn.conn_type_id'            # type of a link connection
    f_link_chain1 = '_struct_conn.ptnr1_label_asym_id'   # asymmetric unit of 1st link component
    f_link_chain2 = '_struct_conn.ptnr2_label_asym_id'   # asymmetric unit of 2nd link component
    f_link_seqid1 = '_struct_conn.ptnr1_label_seq_id'    # sequence id of 1st link component
    f_link_seqid2 = '_struct_conn.ptnr2_label_seq_id'    # sequence id of 2nd link component
    f_link_resid1 = '_struct_conn.ptnr1_label_comp_id'   # residue (cif) of 1st link component
    f_link_resid2 = '_struct_conn.ptnr2_label_comp_id'   # residue (cif) of 2nd link component
    f_link_resid1_au = '_struct_conn.ptnr1_auth_comp_id' # residue (author numeration) of 1st link component
    f_link_resid2_au = '_struct_conn.ptnr2_auth_comp_id' # residue (author numeration) of 2nd link component
    f_link_atom1 = '_struct_conn.ptnr1_label_atom_id'    # atom of 1st link component
    f_link_atom2 = '_struct_conn.ptnr2_label_atom_id'    # atom of 2nd link component
    f_link_list = [f_link_type, f_link_chain1, f_link_chain2, f_link_seqid1, f_link_seqid2,
                   f_link_resid1, f_link_resid2, f_link_resid1_au, f_link_resid2_au,
                   f_link_atom1, f_link_atom2]

    #features of structure
    f_struct_chain = '_atom_site.label_asym_id' # asymmetric unit
    f_struct_seqid = '_atom_site.label_seq_id'  # sequence id
    f_struct_resid = '_atom_site.label_comp_id' # residue
    f_struct_atom = '_atom_site.label_atom_id'  # atom
    f_struct_x = '_atom_site.Cartn_x'           # x coordinate
    f_struct_y = '_atom_site.Cartn_y'           # y coordinate
    f_struct_z = '_atom_site.Cartn_z'           # z coorindate
    f_struct_list = [f_struct_chain, f_struct_seqid, f_struct_resid, 
                     f_struct_atom, f_struct_x, f_struct_y, f_struct_z]

    ### fill the chain class information
    entity_data = iterate_with_features(f_entities, f_chains)
    chain_dict = {}
    for entity, chain in set(iter(entity_data)):
        if chain in chain_dict: 
            if chain_dict[chain] != entity:
                print('One chain belongs to many entities?!')
        else:
            chain_dict[chain] = entity
    chains_all = list(chain_dict.keys())
    chain_data = {chain: Chain(chain) for chain in chains_all}
    cross_chain_bridges = []

    # fill Chain objects with aminoacid sequence
    seq_data = iterate_with_features(f_seq_list, cifdict) # creating generator
    for q_chain, q_resid, q_resid_pdb, q_seqid in iter(seq_data):
        chain_data[q_chain].add_residue(q_resid)
        if q_resid_pdb == '?':
            chain_data[q_chain].add_missing(int(q_seqid), q_resid)

    # fill Chain objects with ssbonds and other links
    link_data = iterate_with_features(f_link_list, cifdict) # creating generator
    metal_ions = {} # container for eventual metal-coordainated and ionic bonds
    for (l_type, l_chain1, l_chain2, l_seqid1, l_seqid2, l_resid1, l_resid2,\
         l_resid1_au, l_resid2_au, l_atom1, l_atom2) in iter(link_data):

        # searching for all covalent bonds (disulfides too)
        if l_type in ['disulf', 'covale']:
            if l_type == 'disulf':
                bridge_type = 'SS'
            else:
                a1, a2 = sorted((l_atom1[0], l_atom2[0]))
                if a1 == "C":
                    if   a2 == "N": bridge_type = "Amide"
                    elif a2 == "O": bridge_type = "Ester"
                    elif a2 == "S": bridge_type = "Thioester"
                    else:           bridge_type = 'Others'
                else: bridge_type = 'Others'
            # filling containers with covalent link data
            # if chainging please look also 30 lines lower
            if l_chain1 == l_chain2:
                chain_data[l_chain1].add_bridge([bridge_type,
                                                 l_resid1, l_atom1, int(l_seqid1),
                                                 l_resid2, l_atom2, int(l_seqid2)])
            else:
                cross_chain_bridges.append([bridge_type,
                                            l_chain1, l_resid1, l_atom1, int(l_seqid1),
                                            l_chain2, l_resid2, l_atom2, int(l_seqid2)])
        # searching for metal coordinated bonds
        elif l_type == 'metalc': 
            # assuming that one set of variables is from metal
            if l_resid1 not in AMINOACIDS:
                if l_resid2 not in AMINOACIDS:
                    print('What is this link? A double ionic bridge???')
                else:
                    l_chain1, l_chain2 = l_chain2, l_chain1
                    l_seqid1, l_seqid2 = l_seqid2, l_seqid1
                    l_resid1_au, l_resid2_au = l_resid2_au, l_resid1_au
                    l_atom1, l_atom2 = l_atom2, l_atom1
                    l_resid1, l_resid2 = l_resid2, l_resid1
            # assuming that var1 is from aminoacid and var2 is from metal
            metal_id = (l_atom2, l_resid2_au)
            # filling temporary metal container with connected aminoacids
            if metal_id in metal_ions:
                metal_ions[metal_id].append([l_chain1, l_resid1, l_atom1, int(l_seqid1)])
            else:
                metal_ions[metal_id] = [[l_chain1, l_resid1, l_atom1, int(l_seqid1)]]

    # searching for aminoacid1-metal-aminoacid2 bridges
    for metal_id, resid_data in metal_ions.items():
        bridge_type = 'Metal_{}'.format(metal_id[0])
        resid_data = sorted(resid_data, key = lambda x: x[3]) # sorting by sequence id
        if len(resid_data) > 1:
            for l1, l2 in itertools.combinations(resid_data,2):
                l_chain1, l_resid1, l_atom1, l_seqid1 = l1
                l_chain2, l_resid2, l_atom2, l_seqid2 = l2
                # filling containers with aminoacid1-aminoacid2 data
                # if chainging please look also 30 lines higher
                if l_chain1 == l_chain2:
                    chain_data[l_chain1].add_bridge([bridge_type,
                                                     l_resid1, l_atom1, int(l_seqid1),
                                                     l_resid2, l_atom2, int(l_seqid2)])
                else:
                    cross_chain_bridges.append([bridge_type,
                                                l_chain1, l_resid1, l_atom1, int(l_seqid1),
                                                l_chain2, l_resid2, l_atom2, int(l_seqid2)])

    # fill Chain objects with coordinates
    struct_data = iterate_with_features(f_struct_list, cifdict) # creating generator
    for s_chain, s_seqid, s_resid, s_atom, x, y, z in iter(struct_data):
        if s_atom == 'CA':
            chain_data[s_chain].add_coordinate(int(s_seqid), [float(x), float(y), float(z)], s_resid)

    ### final collecting
    result = []
    result_gaps = {}
    result_str_begin = {}
    for chain in chains_all:
        chain_data[chain].clean()  # clean bridges data - do not comment
        chain_data[chain].chain_print(work_dir + outfile)  # save coordinates to .xyz and .pdb file
        # chain_data[chain].commands_print(sys.argv[1])	# print input commands to Wanda's program consistent with files saved line above. Addition of second argument=1 prints as first field the bond type
        result.extend(chain_data[chain].getBridges())
        result_gaps[chain_data[chain].name] = chain_data[chain].getGaps()
        result_str_begin[chain_data[chain].name] = chain_data[chain].find_first()
    return result, result_gaps, result_str_begin



from spyprot import PdbFile
from spyprot.pdb2Wanda import run_pdb2Wanda
from spyprot import MMCIFfile
from spyprot.cif2Wanda import run_cif2Wanda
import tempfile
import os
import pytest

def test_pdb2wanda():
    dir = os.path.join(tempfile.gettempdir(), 'test_pdb2wanda')
    os.makedirs(dir, exist_ok=True)
    pdbid = '6xj0'
    p = PdbFile(dir, pdbid).download()

    if os.path.exists(p):
        (bridges, gaps_dict, str_begin_dict) = run_pdb2Wanda(p,
                                                             dir,
                                                             pdbid)  # return tuple of list of tuples: (bridge type, chain_name, bridge_begin, bridge_end) and dict with gaps
        assert bridges == [('SS', 'A', 121, 151)]
        assert gaps_dict == {'A': ''}
        assert str_begin_dict == {'A': 2}


@pytest.mark.skip(reason="Problem with types....")
def test_cif2wanda():
    dir = os.path.join(tempfile.gettempdir(), 'test_pdb2wanda')
    os.makedirs(dir, exist_ok=True)
    pdbid = '6xj0'
    p = MMCIFfile(dir, pdbid).download()

    if os.path.exists(p):
        (bridges, gaps_dict, str_begin_dict) = run_cif2Wanda(p,
                                                             dir,
                                                             pdbid)  # return tuple of list of tuples: (bridge type, chain_name, bridge_begin, bridge_end) and dict with gaps
        assert bridges == [('SS', 'A', 121, 151)]
        assert gaps_dict == {'A': ''}
        assert str_begin_dict == {'A': 2}



import logging
import os
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from urllib.request import urlopen, Request

logger = logging.getLogger(__name__)


def get_proteins(search_url):
    req = Request(search_url, method='GET')
    proteins = []
    with urlopen(req) as resp:
        data = resp.read().decode('utf-8')
        for line in data.splitlines():
            if len(line) > 4 and not line.startswith('#'):
                cells = line.strip().split(';')
                proteins.append((cells[0], cells[1]))
    return proteins


class ILBSMDatabaseDownloader:
    def __init__(self, search_url, download_urls, directory='out', create_separate_dirs=False):
        self.directory = Path(directory)
        self.setup_download_dir()
        self.proteins = get_proteins(search_url)
        self.download_urls = download_urls
        self.create_separate_dirs = create_separate_dirs

    def download_link(self, download_url, protein):
        file_name = download_url.split('/')[-1]
        if not file_name.find('{0}') >= 0:
            file_name = '{0}_{1}_' + file_name
        directory_dir = os.path.join(self.directory, protein[0], protein[1]) if self.create_separate_dirs else self.directory
        if self.create_separate_dirs and not os.path.exists(directory_dir):
            os.makedirs(directory_dir, exist_ok=True)
        download_path = os.path.join(directory_dir, os.path.basename(file_name.format(protein[0], protein[1])))
        link = download_url.format(protein[0], protein[1])
        with urlopen(link) as image, open(download_path, 'wb') as f:
            f.write(image.read())
        logger.debug('Downloaded %s', link)

    def setup_download_dir(self):
        if not self.directory.exists():
            self.directory.mkdir()

    def get_all(self):
        for url in self.download_urls:
            download = partial(self.download_link, url)
            with Pool(8) as pl:
                pl.map(download, self.proteins)




import filecmp
import os
import tempfile

from spyprot import fetchPDBinfo, getCoordinates
from spyprot.fetchChainInfo import PdbFile, MMCIFfile


def test_download_pdb_files():
    dir = os.path.join(tempfile.gettempdir(), 'test_pdb')
    os.makedirs(dir, exist_ok=True)

    p = PdbFile(dir, "1j85", "A").download()
    p = PdbFile(dir, "1k36", "A").download()
    p = PdbFile(dir, "1k36", "A", 'CA').download()
    p = MMCIFfile(dir, "1k36", "A").download()

    # Download and extract chain 5
    p = PdbFile(dir, "6p5i", "5").download()
    # Download and extract only C3' atoms from chain 5
    p = PdbFile(dir, "6p5i", "5", 'C3\'').download()

    m = MMCIFfile(dir, "1k36", "A").download()
    # Download and extract chain 5
    m = MMCIFfile(dir, "6p5i", "5").download()
    # Download and extract only C3' atoms from chain 5

    MMCIFfile(dir, "3bjx", "A").download()
    MMCIFfile(dir, "1et4", "A").download()

    print('PO CO TEN PLIK? ON MI SIĘ NIE ŚCIĄGA. BARTOSZ')
    # PdbFile(dir, "4v7m", "B3").download()
    # Bundle but not translation
    # PdbFile(dir, "6az3", "2").download()
    PdbFile(dir, "6az3", "2", 'C3\'').download()


def test_cif_parser_vs_xml():
    pdb_chains = ['1j85_A', '6ki1_A', '4v7m_B3', '6az3_B', '6az3_e', '6az3_2_C3\'', '6jgz_B',
                  '6gaz_AA_C3\'', '4wf9_X_C3\'', '6sgb_UY', '6sgb_F5', '6sgb_DH', '6f38_M'
                  ]
    dir = os.path.join(tempfile.gettempdir(), 'test_cif_xml')
    os.makedirs(dir, exist_ok=True)
    from Bio import BiopythonWarning
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', BiopythonWarning)
        for el in pdb_chains:
            print(el)
            pdb = el.split('_')[0]
            chain = el.split('_')[1]
            atom = el.split('_')[2] if len(el.split('_')) > 2 else 'CA'
            m = MMCIFfile(dir, pdb, chain, atom=atom)
            m.download()
            m.get_pdb_data()
            m.save_xyz(dir + '/' + el + '_cif.xyz')
            m.save_pdb(dir + '/' + el + '_cif.pdb')
            c = getCoordinates(pdbcode=pdb, work_dir=dir, atom=atom)
            c.getCalfa(chain=chain, output=dir + '/' + el + '_xml.xyz')
            c.getCalfaPdbFormat(chain=chain, output=dir + '/' + el + '_xml.pdb')
            xyz_cmp = filecmp.cmp(dir + '/' + el + '_cif.xyz', dir + '/' + el + '_xml.xyz', shallow=False)
            if not xyz_cmp:
                print('Problem with xyz file comparison for: ' + el)
            assert xyz_cmp is True
            pdb_cmp = filecmp.cmp(dir + '/' + el + '_cif.pdb', dir + '/' + el + '_xml.pdb', shallow=False)
            if not pdb_cmp:
                print('Problem with pdb file comparison for: ' + el)
            assert pdb_cmp is True

            z = fetchPDBinfo(pdbcode=pdb, chain=chain, atom=atom, work_dir=dir)
            ch_r = z.getOrderedChains()
            ch_l = m.get_chains()
            if ch_l != ch_r:
                print('Problem ' + pdb + ': ' + str(len(ch_l)) + '!=' + str(len(ch_r)) + '\n' + str(ch_l) + '\n' + str(
                    ch_r))
                print('Chains not in XML: ' + str(sorted(list(set(ch_l) - set(ch_r)))))
            assert z.getMissingArray() == m.get_missing_array()
            assert z.getMissing() == m.get_missing()
            assert z.getCAlen() == m.get_ca_len()
            assert z.getSeqOneLetterCode() == m.get_seq_one_letter_code()
            assert z.getSeqLength() == m.get_seq_len()
            assert z.getPdbCreationDate() == m.get_pdb_creation_date()
            pm = m.get_meta_pubmed()
            zpm = z.getPubtitlePubmed()
            for i in range(len(zpm)):
                l = str(zpm[i]).replace(' ', '').lower()
                r = str(pm[i]).replace(' ', '').lower()
                if l != r:
                    print('problem ' + pdb + ' ' + chain + ': ' + l + '!=' + r)
                assert l == r


def test_pdb_parser_vs_xml():
    # Will not work correctly for bundles
    pdb_chains = ['6az3_B', '6az3_e', '1j85_A', '6ki1_A']
    dir = os.path.join(tempfile.gettempdir(), 'test_pdb_xml')
    os.makedirs(dir, exist_ok=True)
    from Bio import BiopythonWarning
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', BiopythonWarning)
        for el in pdb_chains:
            pdb = el.split('_')[0]
            chain = el.split('_')[1]
            m = PdbFile(dir, pdb, chain)
            m.download()
            m.get_pdb_data()
            m.save_xyz(dir + '/' + el + '_cif.xyz')
            m.save_pdb(dir + '/' + el + '_cif.pdb')
            c = getCoordinates(pdbcode=pdb, work_dir=dir)
            c.getCalfa(chain=chain, output=dir + '/' + el + '_xml.xyz')
            c.getCalfaPdbFormat(chain=chain, output=dir + '/' + el + '_xml.pdb')
            assert filecmp.cmp(dir + '/' + el + '_cif.xyz', dir + '/' + el + '_xml.xyz', shallow=False) is True
            assert filecmp.cmp(dir + '/' + el + '_cif.pdb', dir + '/' + el + '_xml.pdb', shallow=False) is True

            z = fetchPDBinfo(pdbcode=pdb, chain=chain, work_dir=dir)
            ch_r = z.getOrderedChains()
            ch_l = m.get_chains()
            if ch_l != ch_r:
                print('Problem ' + pdb + ': ' + str(len(ch_l)) + '!=' + str(len(ch_r)) + '\n' + str(ch_l) + '\n' + str(
                    ch_r))
                print('Chains not in PDB: ' + str(list(set(ch_l) - set(ch_r))))
            assert z.getMissingArray() == m.get_missing_array()
            assert z.getMissing() == m.get_missing()
            assert z.getCAlen() == m.get_ca_len()
            assert z.getSeqOneLetterCode() == m.get_seq_one_letter_code()
            assert z.getSeqLength() == m.get_seq_len()


#test_cif_parser_vs_xml()
# test_pdb_parser_vs_xml()
#
# def test_time_xml():
#     pdb_chains = ['4v7m_B3', '6az3_B', '6az3_e', '1j85_A', '6ki1_A', '6jgz_B',
#                   '6az3_2_C3\'', '6gaz_AA_C3\'', '4wf9_X_C3\'']
#     dir = os.path.join(tempfile.gettempdir(), 'test_cif_xml')
#     os.makedirs(dir, exist_ok=True)
#     from Bio import BiopythonWarning
#     import warnings
#     with warnings.catch_warnings():
#         warnings.simplefilter('ignore', BiopythonWarning)
#         for el in pdb_chains:
#             pdb = el.split('_')[0]
#             chain = el.split('_')[1]
#             atom = el.split('_')[2] if len(el.split('_')) > 2 else 'CA'
#             c = getCoordinates(pdbcode=pdb, work_dir=dir, atom=atom)
#             c.getCalfa(chain=chain, output=dir + '/' + el + '_xml.xyz')
#             c.getCalfaPdbFormat(chain=chain, output=dir + '/' + el + '_xml.pdb')
#             z = fetchPDBinfo(pdbcode=pdb, chain=chain, atom=atom, work_dir=dir)
#             ch_r = z.getOrderedChains()
#             z.getMissingArray()
#             z.getSeqOneLetterCode()
#
#
# def test_time_cif():
#     pdb_chains = ['4v7m_B3', '6az3_B', '6az3_e', '1j85_A', '6ki1_A', '6jgz_B',
#                   '6az3_2_C3\'', '6gaz_AA_C3\'', '4wf9_X_C3\'']
#     dir = os.path.join(tempfile.gettempdir(), 'test_cif_xml')
#     os.makedirs(dir, exist_ok=True)
#     from Bio import BiopythonWarning
#     import warnings
#     with warnings.catch_warnings():
#         warnings.simplefilter('ignore', BiopythonWarning)
#         for el in pdb_chains:
#             pdb = el.split('_')[0]
#             chain = el.split('_')[1]
#             atom = el.split('_')[2] if len(el.split('_')) > 2 else 'CA'
#             m = MMCIFfile(dir, pdb, chain, atom=atom)
#             m.download()
#             m.get_pdb_data()
#             m.save_xyz(dir + '/' + el + '_cif.xyz')
#             m.save_pdb(dir + '/' + el + '_cif.pdb')
#             m.get_missing_array()
#             m.get_seq_one_letter_code()
#             m.get_chains()
#
# def time_it(proc):
#     t0 = time()
#     proc()
#     t = time()-t0
#     print('Done {0} in {1} s.'.format(proc.__name__, round(t, 3)))
#
#
# time_it(test_time_cif)
# time_it(test_time_xml)
# dir = os.path.join(tempfile.gettempdir(), 'test_pdb')
# PdbFile(dir, "4v7m", "B3").download()
# p = PdbFile(dir, "1j85", "A")
# p.download()
# p.get_pdb_creation_date()
# p.get_pdb_data()
# test_download_pdb_files()
# test_pdb_parser_vs_xml()
# test_cif_parser_vs_xml()

# test_download_pdb_files()
# dir = os.path.join(tempfile.gettempdir(), 'test_pdb')
# outf = PdbFile(dir, "4v7m").download()
# print(outf)
#
# PdbFile(dir, "6az3", "2", 'C3\'').download()
#
#
# dir = os.path.join(tempfile.gettempdir(), 'test_pdb')
# #p = PdbFile(dir, "7bl1", "AAA").download()
# p = PdbFile(dir, "7bl1").download()
# dir = os.path.join(tempfile.gettempdir(), 'test_pdb')
# #pdb = '1x0g'
# pdb = '6ki1'
# chain = 'A'
#
# m = MMCIFfile(dir, pdb, chain)
# m.download()
# print(m.get_xyz_list())
# #print(m.get_xyz_list(preserve_seqid=True))
# print(m.get_breaks())
#
# print(m.get_seq_one_letter_code())
# print(m.get_meta_pubmed())
#
# print(m.get_missing())
# broken_res = m.get_breaks()
# missing_array = m.get_missing_array()
# str_len = m.get_ca_len()
# seq_len = m.get_seq_len()
# missing = ", ".join(m.get_missing())
# starting_index = m.get_first_residue_id()
# print(broken_res)
# print(missing_array)
# print(str_len)
# print(seq_len)
# print(missing)
# print(starting_index)
#
# #pdb = '6sj9'
# #m = MMCIFfile(dir, pdb, 'A')
# #m.download()
# #print(m.get_xyz_list())
# #print(m.get_breaks())
#
#
#
# z = fetchPDBinfo(pdbcode=pdb, chain=chain, work_dir=dir)
# print(z.getChains())
# print(z.getPubtitlePubmed())
# print(z.getSeqOneLetterCode())
#
# #z.setChain(chain)
# print(z.getCalfaBreaks())
#
# print(z.getMissing())
# broken_res = z.getCalfaBreaks()
# missing_array = z.getMissingArray()
# sequence = z.getSeqOneLetterCode()
# str_len = int(z.getCAlen())
# seq_len = z.getSeqLength()
# missing = ", ".join(z.getMissing())
# date = z.getPdbCreationDate()
# starting_index = z.getFirstResidueIndex()
# print(broken_res)
# print(missing_array)
# print(str_len)
# print(seq_len)
# print(missing)
# print(starting_index)
#
# c = getCoordinates(pdb, work_dir=dir)
# c.getCalfaPdbFormat(chain=chain, output=dir + '/' + pdb + '__.pdb', preserve_seqid=True)
#
#
# pdbm = PdbMetaData(pdb)
# print(pdbm)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Pawel Rubach 2019
# based on:
# Copyright Michal Jamroz, 2014, jamroz@chem.uw.edu.pl
import datetime
import urllib.request, urllib.error, urllib.parse, gzip, re
import os
from os import path
import urllib.request, json

import re

from requests import Session
from Bio import pairwise2
from Bio.Align import substitution_matrices

REFRESH_FILE_INTERVAL = 7 * 24 * 3600  # Refresh every week

PDB_CHAIN_ENZYME = ["ftp://ftp.ebi.ac.uk/pub/databases/msd/sifts/flatfiles/tsv/pdb_chain_enzyme.tsv.gz",
                    "pdb_chain_enzyme.tsv.gz"]
ENZYME_DAT = ["https://ftp.expasy.org/databases/enzyme/enzyme.dat", "enzyme.dat"]

PDB_CHAIN_PFAM = ["ftp://ftp.ebi.ac.uk/pub/databases/msd/sifts/flatfiles/tsv/pdb_chain_pfam.tsv.gz",
                  "pdb_chain_pfam.tsv.gz"]
PFAM_DESC = ["http://pfam.xfam.org/families?output=text", "pdb_chain_pfam_desc"]

DEFAULT_DATA_FILE_PATH = path.join(path.expanduser("~"), ".local", "spyprot")

PDBe = "https://www.ebi.ac.uk/pdbe/api/mappings/uniprot_to_pfam/"

Uni = "https://www.ebi.ac.uk/pdbe/api/mappings/uniprot_segments/"


class AnnotationBase:
    '''
       Base class for retrieving annotations for a given PDBID and Chain

       Annotations are parsed from files downloaded from databases and stored locally in:
       $HOME/.local/spyprot/
    '''

    def __init__(self, files, pdb, chain='A', data_file_path=DEFAULT_DATA_FILE_PATH,
                 refresh_file_interval=REFRESH_FILE_INTERVAL):
        self.data_file_path = data_file_path
        if not path.exists(data_file_path):
            os.makedirs(data_file_path)
        self.refresh_file_interval = refresh_file_interval
        self.first_file = self.download_if_not_exist(files[0])
        self.sec_file = self.download_if_not_exist(files[1])
        self.pdb = pdb.lower()
        self.chain = chain
        self.results = []
        self.d = []
        self.identifiers = []

    def get(self):
        return self.results

    @staticmethod
    def touch(path):
        with open(path, 'a'):
            os.utime(path, None)

    def download_if_not_exist(self, file_locs):
        enz_file = path.join(self.data_file_path, file_locs[1])
        if not path.isfile(enz_file) or (not path.isfile(enz_file + ".flg") and path.isfile(
                enz_file) and datetime.datetime.now().timestamp() - os.stat(
            enz_file).st_mtime > self.refresh_file_interval):
            print('downloading file: ' + enz_file + " from: " + file_locs[0])
            AnnotationBase.touch(enz_file + ".flg")
            try:
                f = urllib.request.urlopen(file_locs[0])
                fw = open(enz_file + "_NEW", "wb")
                fw.write(f.read())
                fw.close()
                if path.isfile(enz_file):
                    os.rename(enz_file, enz_file + "_OLD")
                os.rename(enz_file + "_NEW", enz_file)
                if path.isfile(enz_file + "_OLD"):
                    os.remove(enz_file + "_OLD")
            except Exception as e:
                print("Problem updating file: " + enz_file + " leaving old one" + str(e))
            if path.isfile(enz_file + ".flg"): os.remove(enz_file + ".flg")
        return enz_file


class ECAnnotation(AnnotationBase):
    '''Retrieve EC annotations for a given PDBID and Chain

       Annotations are parsed from files downloaded from expasy and stored locally in:
       $HOME/.local/spyprot/

       Parameters
       ==========
       pdbcode: string
       chain: string

       Return
       ======
       results: list of tuples of EC Code and EC name
       '''

    def __init__(self, pdb, chain='A'):
        super().__init__([PDB_CHAIN_ENZYME, ENZYME_DAT], pdb, chain)
        q = re.compile(r'^' + self.pdb + '\t' + self.chain + '\t(.*)\t(?P<ec>.*)$')

        with gzip.open(self.first_file) as s:
            for line in s.readlines():
                z = q.match(line.decode('utf-8'))
                if z:
                    self.identifiers.append(z.group('ec'))
            for e in self.identifiers:
                o = self._getNameLocal(e)
                if o == "" and not e.endswith("-"):
                    o = self._getName(e)
                self.results.append((e, o))

    def _getName(self, ec):
        o = ""
        try:
            f = urllib.request.urlopen("https://enzyme.expasy.org/EC/" + ec + ".txt")
            q = re.compile(r'^DE.{3}(?P<name>.*)$')
            for line in f.readlines():
                z = q.match(line.decode('utf-8'))
                if z:
                    o = z.group('name')
                    break

        except Exception as e:
            print("Problem getting EC name: " + ec + "\n" + str(e))
        return o

    def _getNameLocal(self, ec):
        if ec.endswith("-"):
            return ""
        o = ""
        try:
            with open(self.sec_file, encoding='utf-8') as f:
                q = re.compile(r'^ID.{3}(' + ec + ')$')
                id_fnd = False
                for line in f.readlines():
                    z = q.match(line)
                    if z and not id_fnd:
                        q = re.compile(r'^DE.{3}(?P<name>.*)$')
                        id_fnd = True
                        continue
                    if z and id_fnd:
                        o = z.group('name')
                        break

        except Exception as e:
            print("Problem getting EC name: " + ec + "\n" + str(e))
        return o


class MyValidationError(Exception):
    def __init__(self, message, *args):
        self.message = message
        super(MyValidationError, self).__init__(message, *args)


class PfamAnnotation(AnnotationBase):
    '''Retrieve PFAM annotations for a given PDBID and Chain

       Annotations are parsed from files downloaded from pfam and stored locally in:
       $HOME/.local/spyprot/

       Parameters
       ==========
       pdbcode: string
       chain: string

       Return
       ======
       results: list of tuples containing attributes: pdbid, chain, pfam_short, pfam_desc, pfam accession code, knot_type, coverage
       '''

    def __init__(self, pdb, chain='A'):
        super().__init__([PDB_CHAIN_PFAM, PFAM_DESC], pdb, chain)
        self.uniprot = ""
        self.pfam_domains = {}
        self.entangled_regions = {}
        self.coverage = {}
        self.domains = {}
        self.pfam_mappings = {}

        with urllib.request.urlopen(Uni + pdb) as url:
            data = json.loads(url.read().decode())
            self.uniprot = (list(data[pdb]['UniProt'].keys())[0])

        with urllib.request.urlopen(PDBe + self.uniprot) as url:
            data = json.loads(url.read().decode())
            if data[self.uniprot]['Pfam'].keys() != []:
                for i in data[self.uniprot]['Pfam'].keys():
                    self.pfam_domains[i] = data[self.uniprot]['Pfam'][
                        i]  # pfam_domain --> {pfam: {mappinngs, identifier, description}}
                    self.pfam_mappings[i] = [self.pfam_domains[i]['mappings'][0]['unp_start'],
                                             self.pfam_domains[i]['mappings'][0]['unp_end']]

                self.entangled_regions = PfamAnnotation.get_entangled_region(pdb, chain)
                self.coverage, self.domains = PfamAnnotation.get_entangled_domains(self.entangled_regions,
                                                                                   self.pfam_mappings)

                for e in set(self.domains.values()):
                    a, b, c = self._get_names(e)
                    k = self.which_knot(self.domains, e)
                    for i in k:
                        self.results.append((pdb, chain, self.pfam_domains[e]['name'],
                                             self.pfam_domains[e]['description'], e, i, self.coverage[i][e]))

    @staticmethod
    def which_knot(ent_domains, search):
        kn = []
        for knot, domain in ent_domains.items():
            if domain == search:
                kn.append(knot)
        return kn

    def _get_names(self, pfam):
        q = re.compile(r'^' + pfam + '\t(?P<pfam_short>.*)\t(?P<pfam_desc>.*)$')
        with open(self.sec_file, encoding='utf-8') as sd:
            for line in sd:
                z = q.match(line)
                if z:
                    return z.group('pfam_short'), z.group('pfam_desc'), pfam
        return "", "", ""

    @staticmethod
    def get_region(seq1, seq2):
        matrix = substitution_matrices.load("BLOSUM62")
        try:
            alignments = pairwise2.align.localds(seq1, seq2, matrix, -10,
                                                 -0.5)  # czemu akurat te dane --> czy zmiana mialaby wplyw na alignment i przypisaną domene?
            if alignments:
                region = alignments[0][1]
                protein = alignments[0][0]
                i = 0
                j = 0
                if region[0] == "-":
                    while region[i] == "-":
                        i += 1
                    region = region[i:]
                if region[-1] == "-":
                    while region[::-1][j] == "-":
                        j += 1
                    j = j * (-1)
                    region = region[:j]
                return region, i + 1, len(protein) - j * (-1)
        except Exception as e:
            raise MyValidationError("error: problem with the alignment.\n" + str(e))

    @staticmethod
    def fasta2string(fasta):  # fasta[1] takes the first sequence (there can be more than one, eg. structure 2fg7 C)
        out = "".join(fasta[1].split('\n')[1:])
        return out

    @staticmethod
    def nc_cuts_from_knotprot(pdb_id, chain_id):  # find N and C cut of the entanglement region
        candidate_files = ["_nr_chains_knotted_N_C", "_red_chains_knotted_N_C",
                           "_all_chains_knotted_N_C"]  # ??czym sa elementy
        entanglement = {}
        for file in candidate_files:
            with Session() as s:
                site = s.get("https://knotprot.cent.uw.edu.pl/{}".format(file),
                             allow_redirects=False)  # po wydrukowaniu site --> lista struktur i info
                for i in site.text.split('\n'):
                    if pdb_id in i and chain_id in i:  # PDB id;Chain id;Chain Length;{Y,A,T}={Published,Artifact,Not published};{K,S}={Knot,Slipknot};Main knot type (e.g. 31=3.1);N_cut;C_cut;Knots range
                        entanglement_type = "{}{}".format(i.split(';')[4], i.split(';')[5])
                        n_cut = int(i.split(';')[6])
                        c_cut = int(i.split(';')[7])
                        entanglement[entanglement_type] = [n_cut, c_cut]  # e.g. {'K31': [85, 89]}
        if entanglement:
            return entanglement
        else:
            raise MyValidationError("error: PDB id not present in KnotProt API files.")

    @staticmethod
    def seq_structure_from_knotprot(pdb_id,
                                    chain_id):  # returns the sequence of the structure that was used to compute entanglement in KnotProt; it includes only resolved amino acids! no gaps!
        with Session() as s:
            site = s.get("https://knotprot.cent.uw.edu.pl/sequence/{}/{}/".format(pdb_id, chain_id),
                         allow_redirects=False)
            if site.text:
                return site.text
            else:
                raise MyValidationError("error: Couldn't find structure sequence in KnotProt.")

    @staticmethod
    def seq_from_uniprot(
            pdb_id):  # returns fasta of the whole protein, not only the sequence from the structure, sometimes there is more than one sequence
        with Session() as s:
            site = s.get("https://www.uniprot.org/uniprot/?query=database:(type:pdb {})&format=fasta".format(pdb_id),
                         allow_redirects=False)
            if site.text:
                return site.text.split(">")
            else:
                raise MyValidationError("error: Couldn't find protein sequence in UniProt based on PDB id.")

    @staticmethod
    def get_entangled_region(pdb_id, chain_id):
        protein_id = PfamAnnotation.seq_from_uniprot(pdb_id)
        seq_protein = PfamAnnotation.fasta2string(protein_id)  # fasta file
        seq_structure = PfamAnnotation.seq_structure_from_knotprot(pdb_id,
                                                                   chain_id)  # returns the sequence of the structure that was used to compute entanglement in KnotProt
        knotprot_ends = PfamAnnotation.nc_cuts_from_knotprot(pdb_id, chain_id)
        entangled_regions = {}
        for knot in knotprot_ends:
            knotprot_N_end = knotprot_ends[knot][0]
            knotprot_C_end = knotprot_ends[knot][1]
            seq_structure_with_gaps, start_region, stop_region = PfamAnnotation.get_region(seq_protein, seq_structure)
            if seq_structure_with_gaps:
                entangled_region_start = start_region + knotprot_N_end
                entangled_region_end = stop_region - knotprot_C_end
                if not entangled_region_start or not entangled_region_end:
                    entangled_regions[knot] = ["error", "error"]
                else:
                    entangled_regions[knot] = [entangled_region_start, entangled_region_end]
        if not entangled_regions:
            raise MyValidationError("errror: Problem with the localization of the entangled region.")
        return entangled_regions

    @staticmethod
    def get_entangled_domains(dict_knot_ranges,
                              dict_domain_ranges):  # dict_domain_ranges={'PF01746': [22, 221]}; dict_knot_ranges={'K31': [89, 132]}
        coverage = {}
        domains = {}
        for knot in dict_knot_ranges:
            knot_start = dict_knot_ranges[knot][0]
            knot_end = dict_knot_ranges[knot][1]
            coverage[knot] = {}
            domains_joined = ''
            if knot_start != "error":
                knot_length = float(knot_end) - float(knot_start) + 1
                for domain in dict(sorted(dict_domain_ranges.items(), key=lambda item: item[1])):
                    domain_start = dict_domain_ranges[domain][0]
                    domain_end = dict_domain_ranges[domain][1]
                    if domain_start >= knot_start and domain_start <= knot_end and domain_end >= knot_end:
                        perc = (knot_end - domain_start + 1) / knot_length * 100
                        coverage[knot][domain] = round(perc, 1)
                        domains_joined += "-" + domain.split("_")[0]
                    elif domain_start <= knot_start and domain_end >= knot_end:
                        perc = 100
                        coverage[knot][domain] = round(perc, 1)
                        domains_joined += "-" + domain.split("_")[0]
                    elif domain_start <= knot_start and domain_end >= knot_start:
                        perc = (domain_end - knot_start + 1) / knot_length * 100
                        coverage[knot][domain] = round(perc, 1)
                        domains_joined += "-" + domain.split("_")[0]
                    elif domain_start >= knot_start and domain_end <= knot_end:
                        perc = (domain_end - domain_start + 1) / knot_length * 100
                        coverage[knot][domain] = round(perc, 1)
                        domains_joined += "-" + domain.split("_")[0]
            domains[knot] = domains_joined[1:]
        return coverage, domains



import glob
import os
import csv
import shutil
import tarfile
import urllib.request, urllib.error, urllib.parse
import datetime
import json
import warnings
from os import makedirs, path
from itertools import count, groupby

import requests
from Bio.PDB import MMCIFIO, Select, PDBParser, PDBIO
from Bio.PDB.MMCIFParser import MMCIFParser
from mysolr import Solr
import logging, sys

from requests import HTTPError

from spyprot.common import _gunzip

PDBE_SOLR_URL = "https://www.ebi.ac.uk/pdbe/search/pdb"
UNLIMITED_ROWS = 10000000
DEBUG = False

ALLOWED_HETATM = ['MSE', 'ORN', 'PCA', 'DGL']

SEQ_CODE = {"ALA": 'A',
            "CYS": 'C',
            "ASP": 'D',
            "GLU": 'E',
            "PHE": 'F',
            "GLY": 'G',
            "HIS": 'H',
            "ILE": 'I',
            "LYS": 'K',
            "LEU": 'L',
            "MET": 'M',
            "MSE": 'M',
            "ASN": 'N',
            "PYL": 'O',
            "PRO": 'P',
            "GLN": 'Q',
            "ARG": 'R',
            "SER": 'S',
            "THR": 'T',
            "SEC": 'U',
            "VAL": 'V',
            "TRP": 'W',
            "5HP": 'E',
            "ABA": 'A',
            "AIB": 'A',
            "BMT": 'T',
            "CEA": 'C',
            "CGU": 'E',
            "CME": 'C',
            "CRO": 'X',
            "CSD": 'C',
            "CSO": 'C',
            "CSS": 'C',
            "CSW": 'C',
            "CSX": 'C',
            "CXM": 'M',
            "DAL": 'A',
            "DAR": 'R',
            "DCY": 'C',
            "DGL": 'E',
            "DGN": 'Q',
            "DHI": 'H',
            "DIL": 'I',
            "DIV": 'V',
            "DLE": 'L',
            "DLY": 'K',
            "DPN": 'F',
            "DPR": 'P',
            "DSG": 'N',
            "DSN": 'S',
            "DSP": 'D',
            "DTH": 'T',
            "DTR": 'X',
            "DTY": 'Y',
            "DVA": 'V',
            "FME": 'M',
            "HYP": 'P',
            "KCX": 'K',
            "LLP": 'K',
            "MLE": 'L',
            "MVA": 'V',
            "NLE": 'L',
            "OCS": 'C',
            "ORN": 'A',
            "PCA": 'E',
            "PTR": 'Y',
            "SAR": 'G',
            "SEP": 'S',
            "STY": 'Y',
            "TPO": 'T',
            "TPQ": 'F',
            "TYS": 'Y',
            "TYR": 'Y'}


class SpyprotWarning(Warning):
    """Spyprot warning."""
    pass


class ProteinFile:
    '''
       Base classs for download of PDB/CIF files from PDB Database and filter by chain and filter out HOH
       Supports also PDB Bundles when there are many subchains for a given protein
    '''

    def __init__(self, dir, pdbcode, chain=None, atom='CA', filter_by_atom=None, preserve_seqid=False):
        self.pdbcode = pdbcode.lower()
        self.chain = chain
        self.dir = dir
        self.atom = atom
        self.filter_by_atom = filter_by_atom
        self.preserve_seqid = preserve_seqid
        self.pdbdata = []
        self.crystlen = -1
        self.missing_array = []
        self.missing = []
        self.seq_idx = []
        self.seq_len = -1
        ###
        self.parser = None
        self.structure = None
        self.data_file = None
        self.io = None
        self.ext = ''
        try:
            makedirs(self.dir)
        except OSError as e:
            pass

    def _get_structure(self):
        if not self.structure:
            self.structure = self._get_parser().get_structure(self.pdbcode, self.data_file)
        return self.structure

    def _get_parser(self):
        return self.parser

    @staticmethod
    def _get_file(url, file):
        if not path.isfile(file) or path.getsize(file) == 0:
            response = urllib.request.urlopen(url)
            if file.endswith('gz'):
                with open(file, 'wb') as myfile:
                    myfile.write(response.read())
            else:
                html = response.read().decode("UTF-8")
                with open(file, 'w') as myfile:
                    myfile.write(html)

    def get_pdb_data(self):
        first_resid = -1
        for ch in self._get_structure().get_chains():
            if ch.get_id() == self.chain:
                for residue in ch.get_residues():
                    if self.atom in residue.child_dict:
                        first_resid = self._get_seqid_from_residue(residue)
                        break
        prev_seqid = first_resid - 9999999999999
        for ch in self._get_structure().get_chains():
            if ch.get_id() == self.chain:
                for residue in ch.get_residues():
                    if self.atom in residue.child_dict:
                        # Filter HETATM other than those in ALLOWED_HETATM
                        if not residue.id[0].strip() == '' or (
                                residue.id[0].strip().startswith('H_') and residue.resname not in ALLOWED_HETATM):
                            continue
                        seq_id = self._get_seqid_from_residue(residue)
                        if seq_id == prev_seqid:
                            prev_seqid = seq_id
                            print(prev_seqid)
                            continue
                        elif prev_seqid > 0 and seq_id - first_resid != prev_seqid:
                            warnings.warn(
                                'Gap in residues for %s %s between: %i and %i ' % (
                                self.pdbcode, self.chain, prev_seqid, seq_id), SpyprotWarning,
                            )
                        coords = [residue.child_dict[self.atom].coord[i] for i in
                                  range(len(residue.child_dict[self.atom].coord))]
                        line = [seq_id] + coords
                        if None not in line:
                            if self.preserve_seqid:
                                new_seqid = line[0]
                            else:
                                new_seqid = line[0] - first_resid + 1
                        prev_seqid = new_seqid
                        self.pdbdata.append(
                            [new_seqid] + line[1:] + [residue.resname, residue.child_dict[self.atom].bfactor])
        self.crystlen = len(self.pdbdata)
        if self.crystlen == 0:
            return []
        seq = [x[0] for x in self.pdbdata]
        self.missing = [x for x in range(seq[0], seq[-1] + 1) if x not in seq]
        self.missing_array = self.missing
        self.missing = ProteinFile.convertToRanges(self.missing)
        self.seq_idx = seq
        self.seq_len = seq[-1]
        return self.pdbdata

    # Overridden in subclasses
    def _get_seqid_from_residue(self, residue):
        pass

    def get_ca_len(self):
        return self.crystlen

    def get_seq_len(self):
        return self.seq_len

    def get_missing(self):
        return self.missing

    def get_missing_array(self):
        return self.missing_array

    def get_xyz_list(self):
        if not self.pdbdata:
            self.get_pdb_data()
        return [el[:4] for el in self.pdbdata]

    def save_xyz(self, output):
        o = ''
        for l in self.get_xyz_list():
            o += "%4d %8.3f %8.3f %8.3f\n" % (l[0], float(l[1]), float(l[2]), float(l[3]))
        if output != '':
            with open(output, "w") as f:
                f.write(o)

    def save_pdb(self, output):
        if not self.pdbdata:
            self.get_pdb_data()
        o = ''
        for l in self.pdbdata:
            line = (l[0], l[4], self.chain, l[0], float(l[1]), float(l[2]), float(l[3]), float(l[5]))
            if self.atom == 'CA':
                o += "ATOM%7d  CA%5s%2s%4d%12.3f%8.3f%8.3f  1.00%6.2f           C\n" % line
            else:
                o += "ATOM%7d  C3'%4s%2s%4d%12.3f%8.3f%8.3f  1.00%6.2f           C\n" % line
        if output != '':
            with open(output, "w") as f:
                f.write(o)

    def get_chains(self):
        return [ch.get_id() for ch in self._get_structure().get_chains()]

    def get_breaks(self):
        # check chain breaks
        eps = 4.2 * 4.2
        brk = []
        o = self.get_xyz_list()
        for i in range(1, len(o)):
            i1 = o[i - 1][0]
            i2 = o[i][0]
            x1 = o[i - 1][1]
            y1 = o[i - 1][2]
            z1 = o[i - 1][3]

            x2 = o[i][1]
            y2 = o[i][2]
            z2 = o[i][3]

            d = (x1 - x2) ** 2 + (y2 - y1) ** 2 + (z1 - z2) ** 2
            if d > eps:
                brk.append(i1)
                brk.append(i2)
        return brk

    def get_first_residue_id(self):
        for ch in self._get_structure().get_chains():
            if ch.get_id() == self.chain:
                for residue in ch.get_residues():
                    return residue.get_id()[1]
        return 0

    def get_residue_list(self):
        residues = {}
        for ch in self._get_structure().get_chains():
            if ch.get_id() == self.chain:
                for residue in ch.get_residues():
                    if residue.__dict__['resname'] != 'HOH':
                        residues[residue.get_id()] = residue.__dict__['resname']
        return residues

    def get_seq_one_letter_code(self):
        seq = ''
        k = list(SEQ_CODE.keys())
        for ch in self._get_structure().get_chains():
            if ch.get_id() == self.chain:
                for residue in ch.get_residues():
                    if (residue.id[0].strip() == '' and self.atom in residue.child_dict.keys()) or (
                            residue.id[0].strip().startswith('H_') and residue.resname in ALLOWED_HETATM) and (
                            residue.child_dict[self.atom].altloc.strip() == '' or residue.child_dict[
                        self.atom].altloc.strip() == 'A'):
                        if self.atom == 'CA':
                            if residue.resname in k:
                                seq += SEQ_CODE[residue.resname]
                            else:
                                seq += "X"
                        else:
                            seq += residue.resname
        return seq

    def filter_by_chain(self):
        self.io.set_structure(self._get_structure())
        self.out_file = self.data_file.replace(self.ext, "_" + self.chain + self.ext)
        self.io.save(self.out_file, ChainAndResidueSelect(self.chain))
        del self.io

    def filter_by_atom(self):
        self.io.set_structure(self._get_structure())
        self.out_file = self.out_file.replace(self.ext, "_" + self.atom + self.ext)
        self.io.save(self.out_file, ChainAndAtomSelect(chain=self.chain, atom=self.atom))
        del self.io

    @staticmethod
    def convertToRanges(L):
        G = (list(x) for _, x in groupby(L, lambda x, c=count(): next(c) - x))
        return ["-".join(map(str, (g[0], g[-1])[:len(g)])) for g in G]


class ChainAndResidueSelect(Select):
    def __init__(self, chain, model=0, residue_out="HOH"):
        self.chain = chain
        self.model = model
        self.residue_out = residue_out

    def accept_model(self, model):
        if model is None or (model is not None and model.id == self.model):
            return True
        else:
            return False

    def accept_chain(self, chain):
        if chain is None or (chain is not None and chain.id == self.chain):
            return True
        else:
            return False

    def accept_residue(self, residue):
        if residue.resname != self.residue_out and str(residue._id[0]).strip() == '':
            return True
        else:
            return False


class ChainSelect(Select):
    def __init__(self, chain, model=1, residue_out="HOH"):
        self.chain = chain
        self.model = model
        self.residue_out = residue_out

    def accept_chain(self, chain):
        if chain is None or (chain is not None and chain.id == self.chain):
            return True
        else:
            return False

    def accept_residue(self, residue):
        if residue.resname != self.residue_out and str(residue._id[0]).strip() == '':
            return True
        else:
            return False


class ChainAndAtomSelect(Select):
    def __init__(self, chain, model=1, residue_out="HOH", atom='CA'):
        self.chain = chain
        self.model = model
        self.residue_out = residue_out
        self.atom = atom

    def accept_chain(self, chain):
        if chain is None or (chain is not None and chain.id == self.chain):
            return True
        else:
            return False

    def accept_residue(self, residue):
        if residue.resname != self.residue_out and str(residue._id[0]).strip() == '':
            return True
        else:
            return False

    def accept_atom(self, atom):
        if atom is None or (atom is not None and atom.get_name() == self.atom):
            return True
        else:
            return False


class PdbFile(ProteinFile):
    '''
       Download PDB files from RCSB PDB Database and filter by chain
       Supports also PDB Bundles when there are many subchains for a given protein

       example:
       PdbFile("/tmp", "1j85", "A").download()

       Parameters
       ==========
       path: string - where to store PDB file
       code: string - PDB ID
       chain: string - if empty return full PDB file or list of translated PDB files in case of PDB Bundles

    '''

    def __init__(self, dir, pdbcode, chain=None, atom='CA', filter_by_atom=None, preserve_seqid=False):
        super().__init__(dir=dir, pdbcode=pdbcode, chain=chain, atom=atom, filter_by_atom=filter_by_atom,
                         preserve_seqid=preserve_seqid)
        self.data_file = path.join(self.dir, self.pdbcode + '.pdb')
        self.parser = PDBParser()
        self.io = PDBIO()
        self.ext = '.pdb'

    def download(self):
        try:
            ProteinFile._get_file('https://files.rcsb.org/view/' + self.pdbcode.upper() + '.pdb', self.data_file)
            if self.chain is not None:
                self.filter_by_chain()
        except urllib.error.HTTPError as e:
            warnings.warn(
                "%s %s ..trying to download from PDB Bundle"
                % (self.pdbcode, str(self.chain)), SpyprotWarning,
            )
            response = urllib.request.urlopen(
                'https://files.rcsb.org/pub/pdb/compatible/pdb_bundle/' + self.pdbcode.lower()[
                                                                          1:3] + '/' + self.pdbcode.lower() + '/' + self.pdbcode.lower() + '-pdb-bundle.tar.gz')
            tar = tarfile.open(fileobj=response, mode="r|gz")
            tar.extractall(self.dir)
            tar.close()
            mapFile, mapChain, mapFileChain = self.parsePdbBundleChainIdFile(
                self.dir + '/' + self.pdbcode.lower() + '-chain-id-mapping.txt')
            # chain is set
            if self.chain:
                newChain = mapChain.get(self.chain)
                pdbBundleFile = self.dir + '/' + mapFile.get(self.chain)
                if newChain != self.chain:
                    self.parsePdbAndTranslateChain(pdbBundleFile, self.chain, newChain)
                else:
                    shutil.move(self.dir + '/' + mapFile.get(self.chain), self.data_file)
                    self.filter_by_chain()
                fileList = glob.glob(os.path.join(self.dir, self.pdbcode + "*bundle*.pdb"))
                for filePath in fileList:
                    try:
                        os.remove(filePath)
                    except OSError:
                        warnings.warn("Error while deleting file", SpyprotWarning)
            else:
                self.out_files = []
                file_num = 1
                for file in sorted(mapFileChain.keys()):
                    mapChainFiltered = {key: mapChain[key] for key in mapFileChain.get(file)}
                    self.parsePdbAndTranslateAllChains(self.dir + '/' + file, mapChainFiltered, file_num)
                    file_num = file_num + 1
                self.data_file = self.out_files
        if self.filter_by_atom:
            self.filter_by_atom()
        return self.data_file

    @staticmethod
    def parsePdbBundleChainIdFile(chainFile):
        with open(chainFile, encoding='utf-8') as fp:
            line = fp.readline()
            cnt = 1
            files = []
            mapChain = {}
            mapFileChain = {}
            mapFile = {}
            actualFile = ''
            while line:
                line = line.strip().rstrip()
                cnt += 1
                if line.find('pdb-bundle') >= 0:
                    actualFile = line[:-1]
                    files.append(actualFile)
                    line = fp.readline()
                    continue
                elif actualFile != '' and line != '':
                    mapping = line.split()
                    key = mapping[1].strip()
                    val = mapping[0].strip()
                    mapChain[key] = val
                    mapFile[key] = actualFile
                    if not actualFile in mapFileChain.keys():
                        mapFileChain[actualFile] = []
                    mapFileChain[actualFile].append(key)
                line = fp.readline()
            return mapFile, mapChain, mapFileChain

    # Remapping chain for PDB bundles with many subchains
    def parsePdbAndTranslateChain(self, pdbFileIn, chain, newChain):
        # print chain + '->' + newChain
        self.data_file = self.data_file.replace(".pdb", "_" + self.chain + ".pdb")
        with open(pdbFileIn, "r", encoding='utf-8') as infile, open(self.data_file, "w", encoding='utf-8') as outfile:
            reader = csv.reader(infile)
            for i, line in enumerate(reader):
                if line[0].find('ATOM') == 0 or line[0].find('HETATM') == 0:
                    newLine = list(line[0])
                    if newLine[21] == newChain:
                        if len(chain) > 2:
                            newLine += chain
                            newLine[21] = "%"
                        elif len(chain) > 1:
                            newLine[20] = chain[0]
                            newLine[21] = chain[1]
                        else:
                            newLine[21] = chain
                        outfile.write("".join(newLine) + "\n")
                else:
                    outfile.write(line[0] + "\n")

        # Remapping chain for PDB bundles with many subchains

    def parsePdbAndTranslateAllChains(self, pdbFileIn, mapChainFiltered, file_num):
        # print chain + '->' + newChain
        outf = self.data_file.replace(".pdb", "_bundle_" + str(file_num) + ".pdb")
        self.out_files.append(outf)
        with open(pdbFileIn, "r", encoding='utf-8') as infile, open(outf, "w", encoding='utf-8') as outfile:
            reader = csv.reader(infile)
            for i, line in enumerate(reader):
                if line[0].find('ATOM') == 0 or line[0].find('HETATM') == 0:
                    newLine = list(line[0])
                    for chain in mapChainFiltered.keys():
                        newChain = mapChainFiltered.get(chain)
                        if newLine[21] == newChain:
                            if len(chain) > 2:
                                newLine += chain
                                newLine[21] = "%"
                            elif len(chain) > 1:
                                newLine[20] = chain[0]
                                newLine[21] = chain[1]
                            else:
                                newLine[21] = chain
                            outfile.write("".join(newLine) + "\n")
                else:
                    outfile.write(line[0] + "\n")

    # Unreliable
    # def get_pdb_creation_date(self):
    #    return self.get_parser().header['deposition_date']
    def _get_seqid_from_residue(self, residue):
        return int(residue.get_id()[1])


class MMCIFfile(ProteinFile):
    def __init__(self, dir, pdbcode, chain=None, atom='CA', filter_by_atom=None, preserve_seqid=False):
        super().__init__(dir=dir, pdbcode=pdbcode, chain=chain, atom=atom, filter_by_atom=filter_by_atom,
                         preserve_seqid=preserve_seqid)
        self.data_file = path.join(self.dir, self.pdbcode + ".cif")
        self.parser = MMCIFParser()
        self.io = MMCIFIO()
        self.ext = '.cif'

    def download_only(self):
        try:
            if not path.isfile(self.data_file) or path.getsize(self.data_file) == 0:
                ProteinFile._get_file('http://www.ebi.ac.uk/pdbe/entry-files/download/' + self.pdbcode + '.cif.gz',
                                      self.data_file + '.gz')
                _gunzip(self.data_file + '.gz')
                os.remove(self.data_file + '.gz')
        except Exception as e:
            ProteinFile._get_file('http://www.ebi.ac.uk/pdbe/entry-files/download/' + self.pdbcode + '.cif',
                                  self.data_file)

    def download(self):
        self.download_only()
        if self.chain is not None:
            self.filter_by_chain()
        else:
            self.out_file = self.data_file
        if self.filter_by_atom:
            self.filter_by_atom()
        return self.out_file

    def _get_seqid_from_residue(self, residue):
        return int(self._get_parser()._mmcif_dict["_atom_site.label_seq_id"][
                       residue.child_dict[self.atom].serial_number - 1])

    def get_pdb_creation_date(self):
        # return self._get_parser()._mmcif_dict['_pdbx_database_status.recvd_initial_deposition_date'][0]
        return self.get_par_from_dict('_pdbx_database_status.recvd_initial_deposition_date')

    def get_meta_pubmed(self):
        pubmed_id = self.get_par_from_dict('_citation.pdbx_database_id_PubMed')
        doi = self.get_par_from_dict('_citation.pdbx_database_id_DOI')
        title = self.get_par_from_dict('_citation.title')
        desc = self.get_par_from_dict('_struct.title')
        src = self.get_par_from_dict('_entity_src_gen.pdbx_gene_src_scientific_name')
        key = self.get_par_from_dict('_struct_keywords.pdbx_keywords')
        molecutag = self.get_par_from_dict('_entity.pdbx_description')
        return (doi, pubmed_id, desc, title, src, key, molecutag)

    def get_par_from_dict(self, par):
        # return self._get_parser()._mmcif_dict[par][0] if par in self._get_parser()._mmcif_dict else None
        if par in self._get_parser()._mmcif_dict:
            res = self._get_parser()._mmcif_dict[par]
            if type(res) == str:
                return res
            elif type(res) == list:
                return res[0]
            else:
                return 'coooooooo'
        else:
            return None

    def get_seq_one_letter_code_can(self):
        model_id = -1
        for model in self._get_structure().child_list:
            for ch in model.child_list:
                if ch.id == self.chain:
                    model_id = model.id
        return self.parser._mmcif_dict['_entity_poly.pdbx_seq_one_letter_code'][model_id].replace('\n', '')


class SearchException(Exception):
    pass


class SequenceException(Exception):
    pass


class PDBeSolrSearch:
    def __init__(self):
        self.solr = Solr(PDBE_SOLR_URL, version=4)
        logging.basicConfig(level=logging.INFO, stream=sys.stdout,
                            format='LOG|%(asctime)s|%(levelname)s  %(message)s', datefmt='%d-%b-%Y %H:%M:%S')
        logging.getLogger("requests").setLevel(logging.WARNING)

    @staticmethod
    def join_with_AND(selectors):
        return " AND ".join(
            ["%s:%s" % (k, v) for k, v in selectors]
        )

    @staticmethod
    def join_with_OR(selectors):
        return " OR ".join(
            ["%s:%s" % (k, v) for k, v in selectors]
        )

    def exec_query(self, field_list, query_details):
        try:
            query = {"rows": UNLIMITED_ROWS, "fl": field_list, "q": self.join_with_AND(query_details)}
            response = self.solr.search(**query)
            documents = response.documents
            if DEBUG:
                logging.getLogger().debug("Found %d matching entities in %d entries." % (
                    len(documents), len({rd["pdb_id"] for rd in documents})))
            return documents
        except Exception as e:
            raise SearchException('%s: error getting query response from: %s for: %s\n %s' % (
                self.__class__.__name__, PDBE_SOLR_URL, str(query), str(e)))

    def get(self):
        return self.results


class IdenticalChains(PDBeSolrSearch):
    '''
       Find identical chains to a given one

       example:
       a = IdenticalChains("2jlo",chain="A").get()

       Parameters
       ==========
       pdbcode: string - PDB ID
    '''

    def __init__(self, pdbcode, chain):
        super().__init__()

        self.results = []
        documents = self.exec_query("pdb_id,entity_id,chain_id,assembly_composition", [('pdb_id', pdbcode)])

        for i in range(len(documents)):
            chain_id = documents[i]['chain_id']
            if chain in chain_id:
                self.results = sorted(chain_id)
                break


class SimilarChains(PDBeSolrSearch):
    '''
       Find similar chains to a given one with sequence identity given as parameter

       example:
       a = SimilarChains("2jlo",chain="A",identity=90).get()

       Parameters
       ==========
       pdbcode: string - PDB ID
       chain: string
       seq: string - sequence to compare against - instead od pdbcode and chain
       identity: int - (percentage) of sequence identity
    '''

    def __init__(self, pdb=None, chain='A', seq=None, identity=40):
        super().__init__()
        self.chain = chain
        self.pdb = pdb.lower() if pdb else None
        self.identity = float(identity / 100)
        self.identifiers = []
        self.results = []

        try:
            self.seq = seq if seq else self.get_seq()
            self.get_similar()
            self.translate_enity_ids_to_chains()
        except (urllib.error.URLError, HTTPError, SequenceException) as he:
            raise SearchException('Problem looking for similar chains to: ' + pdb + ' ' + chain + ': ' + str(he))

    def get_seq(self):
        documents = self.exec_query("pdb_id,entity_id,chain_id,molecule_sequence", [('pdb_id', self.pdb)])
        for i in range(len(documents)):
            chain_id = documents[i]['chain_id']
            if self.chain in chain_id:
                return documents[i]['molecule_sequence']
        raise SequenceException('SimilarChains: Error getting sequence from PDBe for: %s, %s' % (self.pdb, self.chain))

    def get_similar(self):
        query = """
{
  "query": {
    "type": "terminal",
    "service": "sequence",
    "parameters": {
      "evalue_cutoff": 1,
      "identity_cutoff": %s,
      "target": "pdb_protein_sequence",
      "value": "%s"
    }
  },
  "return_type": "polymer_entity",
  "request_options": {
    "pager": {
      "start": 0,
      "rows": 100
    },
    "scoring_strategy": "sequence",
    "sort": [
      {
        "sort_by": "score",
        "direction": "desc"
      }
    ]
  }
}               
        """ % (self.identity, self.seq)
        url = "https://search.rcsb.org/rcsbsearch/v1/query?json={0}".format(urllib.parse.quote(query))
        try:
            response = requests.get(url)
            if response.status_code == 204 and response.text == '':
                return
            if response.status_code != 200:
                raise SequenceException(response.text)
            result = response.json()
            if 'result_set' in result:
                for el in result['result_set']:
                    self.identifiers.append(el['identifier'])
        except Exception as er:
            raise SearchException('SimilarChains: Error in response from RCSB search for: %s, %s URL:\n%s\n%s' % (
                self.pdb, self.chain, urllib.parse.unquote(url), str(er)))

    def translate_enity_ids_to_chains(self):
        if not self.identifiers:
            return
        ident_list = self.identifiers[0]
        for entr_ent in self.identifiers[1:]:
            ident_list += ' OR ' + entr_ent.lower()
        documents = self.exec_query("pdb_id,entity_id,chain_id,assembly_composition",
                                    [('entry_entity', '(' + ident_list + ')')])
        for i in range(len(documents)):
            pid = documents[i]['pdb_id']
            chain_id = documents[i]['chain_id']
            self.results.append((pid.upper(), sorted(chain_id)[0]))

    def get(self):
        return sorted(self.results)


class ReleasedPDBs(PDBeSolrSearch):
    '''
          Download list of pdb ids released in range from_date - to_date (or in day from_date).

          Return empty list if None found
          example:
          a = ReleasedProteins("2020-10-10", "2020-11-10").get()

          Parameters
          ==========
          from_date: datetime or string in format YYYY-MM-DD
          to_date: datetime or string in format YYYY-MM-DD
          uniq_chains: if True return a list of tuples containing (PDBCode, Chain), else return just PDBCodes.
       '''

    def __init__(self, from_date, to_date='', uniq_chains=True, only_prot=True, only_rna=False):
        super().__init__()
        if type(from_date) is datetime.date:
            from_date = from_date.strftime("%Y-%m-%d")
        if to_date == '':
            to_date = from_date
        if type(to_date) is datetime.date:
            to_date = to_date.strftime("%Y-%m-%d")

        if only_rna:
            molecule_type = 'RNA'
        elif not only_prot:
            molecule_type = '(Protein OR RNA)'
        else:
            molecule_type = 'Protein'

        self.results = []
        documents = self.exec_query("pdb_id,entity_id,chain_id,molecule_type",
                                    [
                                        ('release_date', '[' + from_date + 'T00:00:00Z TO ' + to_date + 'T23:59:59Z]'),
                                        ('molecule_type', molecule_type)
                                    ])
        for i in range(len(documents)):
            pid = documents[i]['pdb_id']
            chain_id = documents[i]['chain_id']
            if uniq_chains:
                self.results.append((pid, sorted(chain_id)[0]))
            else:
                if pid not in self.results:
                    self.results.append(pid)
        self.results = sorted(self.results)


class UniqueChains(PDBeSolrSearch):
    '''
       Find a list of unique chains for a given PDB id

       example:
       a = UniqChains("2jlo").get()

       Parameters
       ==========
       pdbcode: string - PDB ID
    '''

    def __init__(self, pdbcode, only_prot=True, only_rna=False):
        super().__init__()

        if only_rna:
            molecule_type = 'RNA'
        elif not only_prot:
            molecule_type = '(Protein OR RNA)'
        else:
            molecule_type = 'Protein'
        self.results = []
        documents = self.exec_query("pdb_id,entity_id,chain_id,assembly_composition,molecule_type",
                                    [
                                        ('pdb_id', pdbcode),
                                        ('molecule_type', molecule_type)
                                    ])
        for i in range(len(documents)):
            chain_id = documents[i]['chain_id']
            self.results.append(sorted(chain_id)[0])


class PdbMetaData(PDBeSolrSearch):
    '''
       Retrieve meta data for a given structure

       example:
       a = IdenticalChains("2jlo",chain="A").get()

       Parameters
       ==========
       pdbcode: string - PDB ID
    '''

    def __init__(self, pdbcode):
        super().__init__()

        self.results = []
        documents = self.exec_query(
            "pdb_id,entity_id, chain_id, assembly_composition, deposition_date, number_of_protein_chains, "
            "pubmed_id, citation_doi, citation_title, title, mutation, mutation_type, assembly_composition, enzyme_systematic_name, enzyme_name, molecule_sequence",
            [('pdb_id', pdbcode)])

        for i in range(len(documents)):
            chain_id = documents[i]['chain_id']
            self.results = sorted(chain_id)
            # break

            # doi = doi[0].text if doi else None
            # molecutag = molecutag[0].text if molecutag else None
            # key = key[0].text.capitalize() if key else None
            # src = src[0].text if src else None
            # pubmed = pubmed[0].text if pubmed else None
            # title = title[0].text if title else None
            # desc = desc[0].text.capitalize() if desc else None
            #
            # return (doi, pubmed, desc, title, src, key, molecutag)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import urllib.error
import urllib.request
from os import path
from re import compile
from spyprot import Command
import os

DEFAULT_DATA_FILE_PATH = path.join(path.expanduser("~"), ".local", "spyprot")
hcath_tmp = DEFAULT_DATA_FILE_PATH + '/cath-domain-description-file.txt_tmp'
hcath = DEFAULT_DATA_FILE_PATH + '/cath-domain-description-file.txt'
cath_uri = \
    'http://download.cathdb.info/cath/releases/latest-release/cath-classification-data/cath-domain-description-file.txt'


# egrep -v "^NAME|^DLEN|^S|^VER|^NSE|^FOR|^DSEQ" CathDomainDescriptionFile_org >CathDomainDescriptionFile


def getCath(pdbcode, chain):
    '''Retrieve CATH annotations for a given PDBID and Chain
       Supported attributes:
       CLASS
       CATHCODE
       DOMAIN
       ARCH
       TOPOL
       HOMOL

       Annotations are parsed from a cath-domain-description file downloaded from cathdb and stored locally in:
       $HOME/.local/spyprot/

       Parameters
       ==========
       pdbcode: string
       chain: string

       Return
       ======
       results: list of CATH annotations
       '''
    if not path.isfile(hcath):
        if not path.exists(DEFAULT_DATA_FILE_PATH):
            os.makedirs(DEFAULT_DATA_FILE_PATH)
        urllib.request.urlretrieve(cath_uri, hcath_tmp)
        cmd = Command('egrep -v "^NAME|^DLEN|^S|^VER|^NSE|^FOR|^DSEQ" ' + hcath_tmp + '>' + hcath)
        cmd.run(1000)
        os.remove(hcath_tmp)
    pdbcode = pdbcode.lower()
    chain = chain
    record = compile(r'^DOMAIN    ' + pdbcode + chain)
    results = []
    with open(hcath, 'r', encoding='utf-8') as cddf:
        found = False
        for line in cddf:
            if record.match(line):
                found = True
                recorddd = {'pdbcode': pdbcode, 'chain': chain}
            elif 'ENDSEG' in line:
                if found:
                    results.append(recorddd)
                found = False
            if found:
                key = line[:10].strip()
                if key not in recorddd:  # only one line..
                    recorddd[key] = line[10:].strip()
    return results


if __name__ == "__main__":
    d = getCath("1j85", "A")
    print(d)
